@@ -1,5 +1,5 @@
 /* dirname.c -- return all but the last element in a path
-   Copyright (C) 1990 Free Software Foundation, Inc.
+   Copyright (C) 1990, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,22 +15,22 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
+#if HAVE_CONFIG_H
+# include <config.h>
 #endif
 
 #ifdef STDC_HEADERS
-#include <stdlib.h>
+# include <stdlib.h>
 #else
 char *malloc ();
 #endif
-#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
-#include <string.h>
+#if defined STDC_HEADERS || defined HAVE_STRING_H
+# include <string.h>
 #else
-#include <strings.h>
-#ifndef strrchr
-#define strrchr rindex
-#endif
+# include <strings.h>
+# ifndef strrchr
+#  define strrchr rindex
+# endif
 #endif
 
 /* Return the leading directories part of PATH,
