@@ -1,5 +1,5 @@
 /* mountlist.h -- declarations for list of mounted filesystems
-   Copyright (C) 1991, 1992 Free Software Foundation, Inc.
+   Copyright (C) 1991, 1992, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,11 +22,26 @@ struct mount_entry
   char *me_mountdir;		/* Mount point directory pathname. */
   char *me_type;		/* "nfs", "4.2", etc. */
   dev_t me_dev;			/* Device number of me_mountdir. */
+  unsigned int me_dummy : 1;	/* Nonzero for dummy filesystems. */
+  unsigned int me_remote : 1;	/* Nonzero for remote fileystems. */
   struct mount_entry *me_next;
 };
 
-#if __STDC__
-struct mount_entry *read_filesystem_list (int need_fs_type, int all_fs);
-#else
-struct mount_entry *read_filesystem_list ();
+#ifndef PARAMS
+# if defined PROTOTYPES || (defined __STDC__ && __STDC__)
+#  define PARAMS(Args) Args
+# else
+#  define PARAMS(Args) ()
+# endif
+#endif
+
+struct mount_entry *read_filesystem_list PARAMS ((int need_fs_type));
+
+#ifndef ME_DUMMY
+# define ME_DUMMY(fs_name, fs_type) \
+    (!strcmp (fs_type, "auto") || !strcmp (fs_type, "ignore"))
+#endif
+
+#ifndef ME_REMOTE
+# define ME_REMOTE(fs_name, fs_type) (strchr (fs_name, ':') != 0)
 #endif
