@@ -1,5 +1,5 @@
 /* chmod -- change permission modes of files
-   Copyright (C) 89, 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,13 +15,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-/* Options:
-   -R	Recursively change modes of directory contents.
-   -c	Verbosely describe only files whose modes actually change.
-   -f	Do not print error messages about files.
-   -v	Verbosely describe changed modes.
-
-   David MacKenzie <djm@gnu.ai.mit.edu> */
+/* Written by David MacKenzie <djm@gnu.ai.mit.edu> */
 
 #include <config.h>
 #include <stdio.h>
@@ -30,17 +24,35 @@
 
 #include "modechange.h"
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
+#include "savedir.h"
+#include "filemode.h"
+
+enum Change_status
+{
+  CH_SUCCEEDED,
+  CH_FAILED,
+  CH_NO_CHANGE_REQUESTED
+};
+
+enum Verbosity
+{
+  /* Print a message for each file that is processed.  */
+  V_high,
+
+  /* Print a message for each file whose attributes we change.  */
+  V_changes_only,
+
+  /* Do not be verbose.  This is the default. */
+  V_off
+};
 
-void mode_string ();
-char *savedir ();
 void strip_trailing_slashes ();
-char *xmalloc ();
-char *xrealloc ();
 
-static int change_dir_mode __P ((const char *dir,
-				 const struct mode_change *changes,
-				 const struct stat *statp));
+static int change_dir_mode PARAMS ((const char *dir,
+				    const struct mode_change *changes,
+				    const struct stat *statp));
 
 /* The name the program was run with. */
 char *program_name;
@@ -51,11 +63,12 @@ static int recurse;
 /* If nonzero, force silence (no error messages). */
 static int force_silent;
 
-/* If nonzero, describe the modes we set. */
-static int verbose;
+/* Level of verbosity.  */
+static enum Verbosity verbosity = V_off;
 
-/* If nonzero, describe only modes that change. */
-static int changes_only;
+/* The argument to the --reference option.  Use the owner and group IDs
+   of this file.  This file must exist.  */
+static char *reference_file;
 
 /* If nonzero, display usage information and exit.  */
 static int show_help;
@@ -69,28 +82,40 @@ static struct option const long_options[] =
   {"changes", no_argument, 0, 'c'},
   {"silent", no_argument, 0, 'f'},
   {"quiet", no_argument, 0, 'f'},
+  {"reference", required_argument, 0, CHAR_MAX + 1},
   {"verbose", no_argument, 0, 'v'},
   {"help", no_argument, &show_help, 1},
   {"version", no_argument, &show_version, 1},
   {0, 0, 0, 0}
 };
 
-/* Tell the user the mode MODE that file FILE has been set to;
-   if CHANGED is zero, FILE had that mode already. */
+/* Tell the user how/if the MODE of FILE has been changed.
+   CHANGED describes what (if anything) has happened. */
 
 static void
-describe_change (const char *file, short unsigned int mode, int changed)
+describe_change (const char *file, short unsigned int mode,
+		 enum Change_status changed)
 {
   char perms[11];		/* "-rwxrwxrwx" ls-style modes. */
+  const char *fmt;
 
   mode_string (mode, perms);
   perms[10] = '\0';		/* `mode_string' does not null terminate. */
-  if (changed)
-    printf (_("mode of %s changed to %04o (%s)\n"),
-	    file, mode & 07777, &perms[1]);
-  else
-    printf (_("mode of %s retained as %04o (%s)\n"),
-	    file, mode & 07777, &perms[1]);
+  switch (changed)
+    {
+    case CH_SUCCEEDED:
+      fmt = _("mode of %s changed to %04o (%s)\n");
+      break;
+    case CH_FAILED:
+      fmt = _("failed to change mode of %s to %04o (%s)\n");
+      break;
+    case CH_NO_CHANGE_REQUESTED:
+      fmt = _("mode of %s retained as %04o (%s)\n");
+      break;
+    default:
+      abort ();
+    }
+  printf (fmt, file, mode & 07777, &perms[1]);
 }
 
 /* Change the mode of FILE according to the list of operations CHANGES.
@@ -131,22 +156,20 @@ change_file_mode (const char *file, const struct mode_change *changes,
 
   if (newmode != (file_stats.st_mode & 07777))
     {
-      if (verbose)
-	describe_change (file, newmode, 1);
-      if (chmod (file, (int) newmode) == 0)
-	{
-	  if (changes_only)
-	    describe_change (file, newmode, 1);
-	}
-      else
+      int fail = chmod (file, (int) newmode);
+
+      if (verbosity == V_high || (verbosity == V_changes_only && !fail))
+	describe_change (file, newmode, (fail ? CH_FAILED : CH_SUCCEEDED));
+
+      if (fail)
 	{
 	  if (force_silent == 0)
 	    error (0, errno, "%s", file);
 	  errors = 1;
 	}
     }
-  else if (verbose && changes_only == 0)
-    describe_change (file, newmode, 0);
+  else if (verbosity == V_high)
+    describe_change (file, newmode, CH_NO_CHANGE_REQUESTED);
 
   if (recurse && S_ISDIR (file_stats.st_mode))
     errors |= change_dir_mode (file, changes, &file_stats);
@@ -170,7 +193,7 @@ change_dir_mode (const char *dir, const struct mode_change *changes,
   int errors = 0;
 
   errno = 0;
-  name_space = savedir (dir, statp->st_size);
+  name_space = savedir (dir, (unsigned int) statp->st_size);
   if (name_space == NULL)
     {
       if (errno)
@@ -217,13 +240,15 @@ usage (int status)
       printf (_("\
 Usage: %s [OPTION]... MODE[,MODE]... FILE...\n\
   or:  %s [OPTION]... OCTAL_MODE FILE...\n\
+  or:  %s [OPTION]... --reference=RFILE FILE...\n\
 "),
-	      program_name, program_name);
+	      program_name, program_name, program_name);
       printf (_("\
 \n\
   -c, --changes           like verbose but report only when a change is made\n\
   -f, --silent, --quiet   suppress most error messages\n\
   -v, --verbose           output a diagnostic for every file processed\n\
+      --reference=RFILE   use RFILE's mode instead of MODE values\n\
   -R, --recursive         change files and directories recursively\n\
       --help              display this help and exit\n\
       --version           output version information and exit\n\
@@ -231,7 +256,8 @@ Usage: %s [OPTION]... MODE[,MODE]... FILE...\n\
 Each MODE is one or more of the letters ugoa, one of the symbols +-= and\n\
 one or more of the letters rwxXstugo.\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -253,15 +279,14 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  recurse = force_silent = verbose = changes_only = 0;
+  recurse = force_silent = 0;
 
   while (1)
     {
       thisind = optind ? optind : 1;
 
-      c = getopt_long (argc, argv, "RcfvrwxXstugoa,+-=", long_options,
-		       (int *) 0);
-      if (c == EOF)
+      c = getopt_long (argc, argv, "RcfvrwxXstugoa,+-=", long_options, NULL);
+      if (c == -1)
 	break;
 
       switch (c)
@@ -286,17 +311,20 @@ main (int argc, char **argv)
 	    error (1, 0, _("invalid mode"));
 	  modeind = thisind;
 	  break;
+	case CHAR_MAX + 1:
+	  reference_file = optarg;
+	  break;
 	case 'R':
 	  recurse = 1;
 	  break;
 	case 'c':
-	  changes_only = 1;
+	  verbosity = V_changes_only;
 	  break;
 	case 'f':
 	  force_silent = 1;
 	  break;
 	case 'v':
-	  verbose = 1;
+	  verbosity = V_high;
 	  break;
 	default:
 	  usage (1);
@@ -306,13 +334,14 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("chmod (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
   if (show_help)
     usage (0);
 
-  if (modeind == 0)
+  if (modeind == 0 && reference_file == NULL)
     modeind = optind++;
 
   if (optind >= argc)
@@ -321,12 +350,15 @@ main (int argc, char **argv)
       usage (1);
     }
 
-  changes = mode_compile (argv[modeind],
-			  MODE_MASK_EQUALS | MODE_MASK_PLUS | MODE_MASK_MINUS);
+  changes = (reference_file ? mode_create_from_ref (reference_file)
+	     : mode_compile (argv[modeind], MODE_MASK_ALL));
+
   if (changes == MODE_INVALID)
     error (1, 0, _("invalid mode"));
   else if (changes == MODE_MEMORY_EXHAUSTED)
     error (1, 0, _("virtual memory exhausted"));
+  else if (changes == MODE_BAD_REFERENCE)
+    error (1, errno, "%s", reference_file);
 
   for (; optind < argc; ++optind)
     {
@@ -334,5 +366,7 @@ main (int argc, char **argv)
       errors |= change_file_mode (argv[optind], changes, 1);
     }
 
+  if (verbosity != V_off)
+    close_stdout ();
   exit (errors);
 }
