@@ -1,5 +1,5 @@
 /* chown -- change user and group ownership of files
-   Copyright (C) 89, 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,7 +36,10 @@
 #include <getopt.h>
 
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
+#include "savedir.h"
+#include "lchown.h"
 
 #ifndef _POSIX_VERSION
 struct passwd *getpwnam ();
@@ -44,31 +47,42 @@ struct group *getgrnam ();
 struct group *getgrgid ();
 #endif
 
-#ifndef HAVE_ENDPWENT
+#if ! HAVE_ENDPWENT
 # define endpwent() ((void) 0)
 #endif
 
-#ifdef HAVE_LCHOWN
-# define LCHOWN(FILE, OWNER, GROUP) lchown (FILE, OWNER, GROUP)
-#else
-# define LCHOWN(FILE, OWNER, GROUP) 1
-#endif
-
-char *savedir ();
 char *parse_user_spec ();
 void strip_trailing_slashes ();
-char *xmalloc ();
-char *xrealloc ();
 
-static int change_dir_owner __P ((char *dir, uid_t user, gid_t group,
-				  struct stat *statp));
+enum Change_status
+{
+  CH_NOT_APPLIED,
+  CH_SUCCEEDED,
+  CH_FAILED,
+  CH_NO_CHANGE_REQUESTED
+};
+
+enum Verbosity
+{
+  /* Print a message for each file that is processed.  */
+  V_high,
+
+  /* Print a message for each file whose attributes we change.  */
+  V_changes_only,
+
+  /* Do not be verbose.  This is the default. */
+  V_off
+};
+
+static int change_dir_owner PARAMS ((const char *dir, uid_t user, gid_t group,
+				     struct stat *statp));
 
 /* The name the program was run with. */
 char *program_name;
 
 /* If nonzero, and the systems has support for it, change the ownership
    of symbolic links rather than any files they point to.  */
-static int change_symlinks;
+static int change_symlinks = 1;
 
 /* If nonzero, change the ownership of directories recursively. */
 static int recurse;
@@ -76,11 +90,8 @@ static int recurse;
 /* If nonzero, force silence (no error messages). */
 static int force_silent;
 
-/* If nonzero, describe the files we process. */
-static int verbose;
-
-/* If nonzero, describe only owners or groups that change. */
-static int changes_only;
+/* Level of verbosity.  */
+static enum Verbosity verbosity = V_off;
 
 /* The name of the user to which ownership of the files is being given. */
 static char *username;
@@ -88,6 +99,10 @@ static char *username;
 /* The name of the group to which ownership of the files is being given. */
 static char *groupname;
 
+/* The argument to the --reference option.  Use the owner and group IDs
+   of this file.  This file must exist.  */
+static char *reference_file;
+
 /* If nonzero, display usage information and exit.  */
 static int show_help;
 
@@ -98,25 +113,47 @@ static struct option const long_options[] =
 {
   {"recursive", no_argument, 0, 'R'},
   {"changes", no_argument, 0, 'c'},
+  {"dereference", no_argument, 0, CHAR_MAX + 2},
   {"no-dereference", no_argument, 0, 'h'},
   {"quiet", no_argument, 0, 'f'},
   {"silent", no_argument, 0, 'f'},
+  {"reference", required_argument, 0, CHAR_MAX + 1},
   {"verbose", no_argument, 0, 'v'},
   {"help", no_argument, &show_help, 1},
   {"version", no_argument, &show_version, 1},
   {0, 0, 0, 0}
 };
 
-/* Tell the user the user and group names to which ownership of FILE
-   has been given; if CHANGED is zero, FILE had those owners already. */
+/* Tell the user how/if the user and group of FILE have been changed.
+   CHANGED describes what (if anything) has happened. */
 
 static void
-describe_change (char *file, int changed)
+describe_change (const char *file, enum Change_status changed)
 {
-  if (changed)
-    printf (_("owner of %s changed to "), file);
-  else
-    printf (_("owner of %s retained as "), file);
+  const char *fmt;
+
+  if (changed == CH_NOT_APPLIED)
+    {
+      printf (_("neither symbolic link %s nor referent has been changed\n"),
+	      file);
+      return;
+    }
+
+  switch (changed)
+    {
+    case CH_SUCCEEDED:
+      fmt = _("owner of %s changed to ");
+      break;
+    case CH_FAILED:
+      fmt = _("failed to change owner of %s to ");
+      break;
+    case CH_NO_CHANGE_REQUESTED:
+      fmt = _("owner of %s retained as ");
+      break;
+    default:
+      abort ();
+    }
+  printf (fmt, file);
   if (groupname)
     printf ("%s.%s\n", username, groupname);
   else
@@ -128,7 +165,7 @@ describe_change (char *file, int changed)
    Return 0 if successful, 1 if errors occurred. */
 
 static int
-change_file_owner (char *file, uid_t user, gid_t group)
+change_file_owner (int cmdline_arg, const char *file, uid_t user, gid_t group)
 {
   struct stat file_stats;
   uid_t newuser;
@@ -147,14 +184,31 @@ change_file_owner (char *file, uid_t user, gid_t group)
   if (newuser != file_stats.st_uid || newgroup != file_stats.st_gid)
     {
       int fail;
+      int symlink_changed = 1;
 
-      if (verbose)
-	describe_change (file, 1);
+      if (S_ISLNK (file_stats.st_mode) && change_symlinks)
+	{
+	  fail = lchown (file, newuser, newgroup);
 
-      if (change_symlinks)
-	fail = LCHOWN (file, newuser, newgroup);
+	  /* Ignore the failure if it's due to lack of support (ENOSYS)
+	     and this is not a command line argument.  */
+	  if (!cmdline_arg && fail && errno == ENOSYS)
+	    {
+	      fail = 0;
+	      symlink_changed = 0;
+	    }
+	}
       else
-	fail = chown (file, newuser, newgroup);
+	{
+	  fail = chown (file, newuser, newgroup);
+	}
+
+      if (verbosity == V_high || (verbosity == V_changes_only && !fail))
+	{
+	  enum Change_status ch_status = (! symlink_changed ? CH_NOT_APPLIED
+					  : (fail ? CH_FAILED : CH_SUCCEEDED));
+	  describe_change (file, ch_status);
+	}
 
       if (fail)
 	{
@@ -163,8 +217,10 @@ change_file_owner (char *file, uid_t user, gid_t group)
 	  errors = 1;
 	}
     }
-  else if (verbose && changes_only == 0)
-    describe_change (file, 0);
+  else if (verbosity == V_high)
+    {
+      describe_change (file, CH_NO_CHANGE_REQUESTED);
+    }
 
   if (recurse && S_ISDIR (file_stats.st_mode))
     errors |= change_dir_owner (file, user, group, &file_stats);
@@ -177,7 +233,7 @@ change_file_owner (char *file, uid_t user, gid_t group)
    Return 0 if successful, 1 if errors occurred. */
 
 static int
-change_dir_owner (char *dir, uid_t user, gid_t group, struct stat *statp)
+change_dir_owner (const char *dir, uid_t user, gid_t group, struct stat *statp)
 {
   char *name_space, *namep;
   char *path;			/* Full path of each entry to process. */
@@ -187,7 +243,7 @@ change_dir_owner (char *dir, uid_t user, gid_t group, struct stat *statp)
   int errors = 0;
 
   errno = 0;
-  name_space = savedir (dir, statp->st_size);
+  name_space = savedir (dir, (unsigned int) statp->st_size);
   if (name_space == NULL)
     {
       if (errno)
@@ -216,7 +272,7 @@ change_dir_owner (char *dir, uid_t user, gid_t group, struct stat *statp)
 	  path = xrealloc (path, pathlength);
 	}
       strcpy (path + dirlength, namep);
-      errors |= change_file_owner (path, user, group);
+      errors |= change_file_owner (0, path, user, group);
     }
   free (path);
   free (name_space);
@@ -233,16 +289,22 @@ usage (int status)
     {
       printf (_("\
 Usage: %s [OPTION]... OWNER[.[GROUP]] FILE...\n\
-  or:  %s [OPTION]... .[GROUP] FILE...\n\
+  or:  %s [OPTION]... .GROUP FILE...\n\
+  or:  %s [OPTION]... --reference=RFILE FILE...\n\
 "),
-	      program_name, program_name);
+	      program_name, program_name, program_name);
       printf (_("\
 Change the owner and/or group of each FILE to OWNER and/or GROUP.\n\
 \n\
   -c, --changes          be verbose whenever change occurs\n\
+      --dereference      affect the referent of each symbolic link, rather\n\
+                         than the symbolic link itself\n\
   -h, --no-dereference   affect symbolic links instead of any referenced file\n\
-                         (available only on systems with lchown system call)\n\
+                         (available only on systems that can change the\n\
+                         ownership of a symlink)\n\
   -f, --silent, --quiet  suppress most error messages\n\
+      --reference=RFILE  use the owner and group of RFILE instead of using\n\
+                         explicit OWNER.GROUP values\n\
   -R, --recursive        operate on files and directories recursively\n\
   -v, --verbose          explain what is being done\n\
       --help             display this help and exit\n\
@@ -251,7 +313,8 @@ Change the owner and/or group of each FILE to OWNER and/or GROUP.\n\
 Owner is unchanged if missing.  Group is unchanged if missing, but changed\n\
 to login group if implied by a period.  A colon may replace the period.\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -270,21 +333,25 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  recurse = force_silent = verbose = changes_only = 0;
+  recurse = force_silent = 0;
 
-  while ((optc = getopt_long (argc, argv, "Rcfhv", long_options, (int *) 0))
-	 != EOF)
+  while ((optc = getopt_long (argc, argv, "Rcfhv", long_options, NULL)) != -1)
     {
       switch (optc)
 	{
 	case 0:
 	  break;
+	case CHAR_MAX + 1:
+	  reference_file = optarg;
+	  break;
+	case CHAR_MAX + 2:
+	  change_symlinks = 0;
+	  break;
 	case 'R':
 	  recurse = 1;
 	  break;
 	case 'c':
-	  verbose = 1;
-	  changes_only = 1;
+	  verbosity = V_changes_only;
 	  break;
 	case 'f':
 	  force_silent = 1;
@@ -293,7 +360,7 @@ main (int argc, char **argv)
 	  change_symlinks = 1;
 	  break;
 	case 'v':
-	  verbose = 1;
+	  verbosity = V_high;
 	  break;
 	default:
 	  usage (1);
@@ -303,36 +370,47 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("chown (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
   if (show_help)
     usage (0);
 
-  if (optind >= argc - 1)
+  if (argc - optind + (reference_file ? 1 : 0) <= 1)
     {
       error (0, 0, _("too few arguments"));
       usage (1);
     }
 
-#ifndef HAVE_LCHOWN
-  if (change_symlinks)
+  if (reference_file)
     {
-      error (1, 0, _("--no-dereference (-h) is not supported on this system"));
+      struct stat ref_stats;
+
+      if (stat (reference_file, &ref_stats))
+        error (1, errno, "%s", reference_file);
+
+      user  = ref_stats.st_uid;
+      group = ref_stats.st_gid;
     }
-#endif
+  else
+    {
+      e = parse_user_spec (argv[optind], &user, &group, &username, &groupname);
+      if (e)
+        error (1, 0, "%s: %s", argv[optind], e);
+      if (username == NULL)
+        username = "";
 
-  e = parse_user_spec (argv[optind], &user, &group, &username, &groupname);
-  if (e)
-    error (1, 0, "%s: %s", argv[optind], e);
-  if (username == NULL)
-    username = "";
+      optind++;
+    }
 
-  for (++optind; optind < argc; ++optind)
+  for (; optind < argc; ++optind)
     {
       strip_trailing_slashes (argv[optind]);
-      errors |= change_file_owner (argv[optind], user, group);
+      errors |= change_file_owner (1, argv[optind], user, group);
     }
 
+  if (verbosity != V_off)
+    close_stdout ();
   exit (errors);
 }
