@@ -1,5 +1,5 @@
 /* dd -- convert a file while copying it.
-   Copyright (C) 85, 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 85, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,55 +17,28 @@
 
 /* Written by Paul Rubin, David MacKenzie, and Stuart Kemp. */
 
-/* Options:
-
-   Numbers can be followed by a multiplier:
-   b=512, c=1, k=1024, w=2, xm=number m
-
-   if=FILE			Read from FILE instead of stdin.
-   of=FILE			Write to FILE instead of stdout; don't
-				truncate FILE.
-   ibs=BYTES			Read BYTES bytes at a time.
-   obs=BYTES			Write BYTES bytes at a time.
-   bs=BYTES			Override ibs and obs.
-   cbs=BYTES			Convert BYTES bytes at a time.
-   skip=BLOCKS			Skip BLOCKS ibs-sized blocks at
-				start of input.
-   seek=BLOCKS			Skip BLOCKS obs-sized blocks at
-				start of output.
-   count=BLOCKS			Copy only BLOCKS input blocks.
-   conv=CONVERSION[,CONVERSION...]
-
-   Conversions:
-   ascii			Convert EBCDIC to ASCII.
-   ebcdic			Convert ASCII to EBCDIC.
-   ibm				Convert ASCII to alternate EBCDIC.
-   block			Pad newline-terminated records to size of
-				cbs, replacing newline with trailing spaces.
-   unblock			Replace trailing spaces in cbs-sized block
-				with newline.
-   lcase			Change upper case characters to lower case.
-   ucase			Change lower case characters to upper case.
-   swab				Swap every pair of input bytes.
-				Unlike the Unix dd, this works when an odd
-				number of bytes are read.
-   noerror			Continue after read errors.
-   sync				Pad every input block to size of ibs with
-				trailing NULs. */
-
 #include <config.h>
 #include <stdio.h>
 
 #define SWAB_ALIGN_OFFSET 2
 
+#if HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
 #include <sys/types.h>
 #include <signal.h>
 #include <getopt.h>
 
+#include "human.h"
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
+#include "safe-read.h"
+
+#ifndef SIGINFO
+# define SIGINFO SIGUSR1
+#endif
 
-#define equal(p, q) (strcmp ((p),(q)) == 0)
 #define max(a, b) ((a) > (b) ? (a) : (b))
 #define output_char(c) \
   do { \
@@ -90,28 +63,8 @@
 /* Use separate input and output buffers, and combine partial input blocks. */
 #define C_TWOBUFS 04000
 
-char *xmalloc ();
-int safe_read ();
 int full_write ();
 
-static RETSIGTYPE interrupt_handler __P ((int));
-static int bit_count __P ((register unsigned int i));
-static int parse_integer __P ((char *str));
-static void apply_translations __P ((void));
-static void copy __P ((void));
-static void copy_simple __P ((unsigned char *buf, int nread));
-static void copy_with_block __P ((unsigned char *buf, int nread));
-static void copy_with_unblock __P ((unsigned char *buf, int nread));
-static void parse_conversion __P ((char *str));
-static void print_stats __P ((void));
-static void translate_charset __P ((const unsigned char *new_trans));
-static void quit __P ((int code));
-static void scanargs __P ((int argc, char **argv));
-static void skip __P ((int fdesc, char *file, long int records,
-		       long int blocksize, unsigned char *buf));
-static void usage __P ((int status));
-static void write_output __P ((void));
-
 /* The name this program was run with. */
 char *program_name;
 
@@ -128,22 +81,22 @@ static char *output_file = NULL;
 static int output_fd = 1;
 
 /* The number of bytes in which atomic reads are done. */
-static long input_blocksize = -1;
+static size_t input_blocksize = 0;
 
 /* The number of bytes in which atomic writes are done. */
-static long output_blocksize = -1;
+static size_t output_blocksize = 0;
 
 /* Conversion buffer size, in bytes.  0 prevents conversions. */
-static long conversion_blocksize = 0;
+static size_t conversion_blocksize = 0;
 
 /* Skip this many records of `input_blocksize' bytes before input. */
-static long skip_records = 0;
+static uintmax_t skip_records = 0;
 
 /* Skip this many records of `output_blocksize' bytes before output. */
-static long seek_record = 0;
+static uintmax_t seek_record = 0;
 
-/* Copy only this many records.  <0 means no limit. */
-static int max_records = -1;
+/* Copy only this many records.  The default is effectively infinity.  */
+static uintmax_t max_records = (uintmax_t) -1;
 
 /* Bit vector of conversions to apply. */
 static int conversions_mask = 0;
@@ -152,25 +105,34 @@ static int conversions_mask = 0;
 static int translation_needed = 0;
 
 /* Number of partial blocks written. */
-static unsigned w_partial = 0;
+static uintmax_t w_partial = 0;
 
 /* Number of full blocks written. */
-static unsigned w_full = 0;
+static uintmax_t w_full = 0;
 
 /* Number of partial blocks read. */
-static unsigned r_partial = 0;
+static uintmax_t r_partial = 0;
 
 /* Number of full blocks read. */
-static unsigned r_full = 0;
+static uintmax_t r_full = 0;
 
 /* Records truncated by conv=block. */
-static unsigned r_truncate = 0;
+static uintmax_t r_truncate = 0;
 
 /* Output representation of newline and space characters.
    They change if we're converting to EBCDIC.  */
 static unsigned char newline_character = '\n';
 static unsigned char space_character = ' ';
 
+/* Output buffer. */
+static unsigned char *obuf;
+
+/* Current index into `obuf'. */
+static size_t oc = 0;
+
+/* Index into current line, for `conv=block' and `conv=unblock'.  */
+static size_t col = 0;
+
 struct conversion
 {
   char *convname;
@@ -317,146 +279,456 @@ static struct option const long_options[] =
   {0, 0, 0, 0}
 };
 
-int
-main (int argc, char **argv)
+static void
+usage (int status)
+{
+  if (status != 0)
+    fprintf (stderr, _("Try `%s --help' for more information.\n"),
+	     program_name);
+  else
+    {
+      printf (_("Usage: %s [OPTION]...\n"), program_name);
+      printf (_("\
+Copy a file, converting and formatting according to the options.\n\
+\n\
+  bs=BYTES        force ibs=BYTES and obs=BYTES\n\
+  cbs=BYTES       convert BYTES bytes at a time\n\
+  conv=KEYWORDS   convert the file as per the comma separated keyword list\n\
+  count=BLOCKS    copy only BLOCKS input blocks\n\
+  ibs=BYTES       read BYTES bytes at a time\n\
+  if=FILE         read from FILE instead of stdin\n\
+  obs=BYTES       write BYTES bytes at a time\n\
+  of=FILE         write to FILE instead of stdout\n\
+  seek=BLOCKS     skip BLOCKS obs-sized blocks at start of output\n\
+  skip=BLOCKS     skip BLOCKS ibs-sized blocks at start of input\n\
+      --help      display this help and exit\n\
+      --version   output version information and exit\n\
+\n\
+BYTES may be followed by the following multiplicative suffixes:\n\
+xM M, c 1, w 2, b 512, kD 1000, k 1024, MD 1,000,000, M 1,048,576,\n\
+GD 1,000,000,000, G 1,073,741,824, and so on for T, P, E, Z, Y.\n\
+Each KEYWORD may be:\n\
+\n\
+  ascii     from EBCDIC to ASCII\n\
+  ebcdic    from ASCII to EBCDIC\n\
+  ibm       from ASCII to alternated EBCDIC\n\
+  block     pad newline-terminated records with spaces to cbs-size\n\
+  unblock   replace trailing spaces in cbs-size records with newline\n\
+  lcase     change upper case to lower case\n\
+  notrunc   do not truncate the output file\n\
+  ucase     change lower case to upper case\n\
+  swab      swap every pair of input bytes\n\
+  noerror   continue after read errors\n\
+  sync      pad every input block with NULs to ibs-size\n\
+"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
+    }
+  exit (status);
+}
+
+static void
+translate_charset (const unsigned char *new_trans)
+{
+  int i;
+
+  for (i = 0; i < 256; i++)
+    trans_table[i] = new_trans[trans_table[i]];
+  translation_needed = 1;
+}
+
+/* Return the number of 1 bits in `i'. */
+
+static int
+bit_count (register unsigned int i)
+{
+  register int set_bits;
+
+  for (set_bits = 0; i != 0; set_bits++)
+    i &= i - 1;
+  return set_bits;
+}
+
+static void
+print_stats (void)
+{
+  char buf[2][LONGEST_HUMAN_READABLE + 1];
+  fprintf (stderr, _("%s+%s records in\n"),
+	   human_readable (r_full, buf[0], 1, 1),
+	   human_readable (r_partial, buf[1], 1, 1));
+  fprintf (stderr, _("%s+%s records out\n"),
+	   human_readable (w_full, buf[0], 1, 1),
+	   human_readable (w_partial, buf[1], 1, 1));
+  if (r_truncate > 0)
+    {
+      fprintf (stderr, "%s %s\n",
+	       human_readable (r_truncate, buf[0], 1, 1),
+	       (r_truncate == 1
+		? _("truncated record")
+		: _("truncated records")));
+    }
+}
+
+static void
+cleanup (void)
+{
+  print_stats ();
+  if (close (input_fd) < 0)
+    error (1, errno, "%s", input_file);
+  if (close (output_fd) < 0)
+    error (1, errno, "%s", output_file);
+}
+
+static inline void
+quit (int code)
+{
+  cleanup ();
+  exit (code);
+}
+
+static RETSIGTYPE
+interrupt_handler (int sig)
+{
+#ifdef SA_INTERRUPT
+  struct sigaction sigact;
+
+  sigact.sa_handler = SIG_DFL;
+  sigemptyset (&sigact.sa_mask);
+  sigact.sa_flags = 0;
+  sigaction (sig, &sigact, NULL);
+#else				/* !SA_INTERRUPT */
+  signal (sig, SIG_DFL);
+#endif				/* SA_INTERRUPT */
+  cleanup ();
+  kill (getpid (), sig);
+}
+
+static RETSIGTYPE
+siginfo_handler (int sig)
+{
+  print_stats ();
+}
+
+/* Encapsulate portability mess of establishing signal handlers.  */
+
+static void
+install_handler (int sig_num, RETSIGTYPE (*sig_handler) (int sig))
 {
 #ifdef _POSIX_VERSION
   struct sigaction sigact;
-#endif /* _POSIX_VERSION */
-  int i;
-
-  program_name = argv[0];
-  setlocale (LC_ALL, "");
-  bindtextdomain (PACKAGE, LOCALEDIR);
-  textdomain (PACKAGE);
-
-  /* Initialize translation table to identity translation. */
-  for (i = 0; i < 256; i++)
-    trans_table[i] = i;
-
-  /* Decode arguments. */
-  scanargs (argc, argv);
-
-  if (show_version)
+  sigaction (sig_num, NULL, &sigact);
+  if (sigact.sa_handler != SIG_IGN)
     {
-      printf ("dd (%s) %s\n", GNU_PACKAGE, VERSION);
-      exit (0);
+      sigact.sa_handler = sig_handler;
+      sigemptyset (&sigact.sa_mask);
+      sigact.sa_flags = 0;
+      sigaction (sig_num, &sigact, NULL);
     }
-
-  if (show_help)
-    usage (0);
-
-  apply_translations ();
-
-  if (input_file != NULL)
-    {
-      input_fd = open (input_file, O_RDONLY);
-      if (input_fd < 0)
-	error (1, errno, "%s", input_file);
-    }
-  else
-    input_file = _("standard input");
-
-  if (input_fd == output_fd)
-    error (1, 0, _("%s is closed"), (input_fd == 0
-				     ? _("standard input")
-				     : _("standard output")));
-
-  if (output_file != NULL)
-    {
-      int omode = O_RDWR | O_CREAT;
-
-      if (seek_record == 0 && !(conversions_mask & C_NOTRUNC))
-	omode |= O_TRUNC;
-      output_fd = open (output_file, omode, 0666);
-      if (output_fd < 0)
-	error (1, errno, "%s", output_file);
-#ifdef HAVE_FTRUNCATE
-      if (seek_record > 0 && !(conversions_mask & C_NOTRUNC))
-	{
-	  if (ftruncate (output_fd, seek_record * output_blocksize) < 0)
-	    error (0, errno, "%s", output_file);
-	}
+#else
+  if (signal (sig_num, SIG_IGN) != SIG_IGN)
+    signal (sig_num, sig_handler);
 #endif
-    }
-  else
-    output_file = _("standard output");
-
-#ifdef _POSIX_VERSION
-  sigaction (SIGINT, NULL, &sigact);
-  if (sigact.sa_handler != SIG_IGN)
-    {
-      sigact.sa_handler = interrupt_handler;
-      sigemptyset (&sigact.sa_mask);
-      sigact.sa_flags = 0;
-      sigaction (SIGINT, &sigact, NULL);
-    }
-  sigaction (SIGPIPE, NULL, &sigact);
-  if (sigact.sa_handler != SIG_IGN)
-    {
-      sigact.sa_handler = interrupt_handler;
-      sigemptyset (&sigact.sa_mask);
-      sigact.sa_flags = 0;
-      sigaction (SIGPIPE, &sigact, NULL);
-    }
-#else				/* !_POSIX_VERSION */
-  if (signal (SIGINT, SIG_IGN) != SIG_IGN)
-    signal (SIGINT, interrupt_handler);
-  if (signal (SIGPIPE, SIG_IGN) != SIG_IGN)
-    signal (SIGPIPE, interrupt_handler);
-#endif				/* !_POSIX_VERSION */
-  copy ();
 }
 
-/* Throw away RECORDS blocks of BLOCKSIZE bytes on file descriptor FDESC,
-   which is open with read permission for FILE.  Store up to BLOCKSIZE
-   bytes of the data at a time in BUF, if necessary. */
+/* Write, then empty, the output buffer `obuf'. */
 
 static void
-skip (int fdesc, char *file, long int records, long int blocksize,
-      unsigned char *buf)
+write_output (void)
 {
-  struct stat stats;
-
-  /* Use fstat instead of checking for errno == ESPIPE because
-     lseek doesn't work on some special files but doesn't return an
-     error, either. */
-  /* FIXME: can this really happen?  What system?  */
-  if (fstat (fdesc, &stats))
+  int nwritten = full_write (output_fd, obuf, output_blocksize);
+  if (nwritten != output_blocksize)
     {
-      error (0, errno, "%s", file);
+      error (0, errno, "%s", output_file);
+      if (nwritten > 0)
+	w_partial++;
       quit (1);
     }
+  else
+    w_full++;
+  oc = 0;
+}
 
-  /* FIXME: why use lseek only on regular files?
-     Better: try lseek and if an error indicates it was an inappropriate
-     operation, fall back on using read.  */
-  if (S_ISREG (stats.st_mode))
+/* Interpret one "conv=..." option. */
+
+static void
+parse_conversion (char *str)
+{
+  char *new;
+  int i;
+
+  do
     {
-      if (lseek (fdesc, records * blocksize, SEEK_SET) < 0)
+      new = strchr (str, ',');
+      if (new != NULL)
+	*new++ = '\0';
+      for (i = 0; conversions[i].convname != NULL; i++)
+	if (STREQ (conversions[i].convname, str))
+	  {
+	    conversions_mask |= conversions[i].conversion;
+	    break;
+	  }
+      if (conversions[i].convname == NULL)
 	{
-	  error (0, errno, "%s", file);
-	  quit (1);
+	  error (0, 0, _("%s: invalid conversion"), str);
+	  usage (1);
+	}
+      str = new;
+  } while (new != NULL);
+}
+
+/* Return the value of STR, interpreted as a non-negative decimal integer,
+   optionally multiplied by various values.
+   Assign nonzero to *INVALID if STR does not represent a number in
+   this format. */
+
+/* FIXME: use xstrtou?[lq] */
+
+static uintmax_t
+parse_integer (char *str, int *invalid)
+{
+  register uintmax_t n = 0;
+  register char *p = str;
+
+  while (ISDIGIT (*p))
+    {
+      uintmax_t n10 = n * 10;
+      int digit = *p - '0';
+      if (! (n10 / 10 == n && n10 <= n10 + digit))
+	{
+	  *invalid = 1;
+	  return 0;
+	}
+      n = n10 + digit;
+      p++;
+    }
+
+  for (;;)
+    {
+      uintmax_t multiplier;
+      int power = 0;
+
+#ifdef lint
+      /* Suppress `used before initialized' warning.  */
+      multiplier = 0;
+#endif
+
+      switch (*p++)
+	{
+	case '\0':
+	  return n;
+	case 'b':
+	  multiplier = 512;
+	  break;
+	case 'c':
+	  continue;
+	case 'E': /* Exa */
+	  power = 6;
+	  break;
+	case 'G': /* Giga */
+	  power = 3;
+	  break;
+	case 'k': /* kilo */
+	  power = 1;
+	  break;
+	case 'M': /* Mega */
+	  power = 2;
+	  break;
+	case 'P': /* Peta */
+	  power = 5;
+	  break;
+	case 'T': /* Tera */
+	  power = 4;
+	  break;
+	case 'w':
+	  multiplier = 2;
+	  break;
+	case 'x':
+	  multiplier = parse_integer (p, invalid);
+	  p = "";
+	  break;
+	case 'Y': /* Yotta */
+	  power = 8;
+	  break;
+	case 'Z': /* Zetta */
+	  power = 7;
+	  break;
+	default:
+	  {
+	    *invalid = 1;
+	    return 0;
+	  }
+	}
+
+      if (power)
+	{
+	  int base = 1024;
+
+	  switch (*p)
+	    {
+	    case 'B': p++; break;
+	    case 'D': p++; base = 1000; break;
+	    }
+
+	  for (multiplier = base;  --power;  multiplier *= base)
+	    if (multiplier * base / base != multiplier)
+	      {
+		*invalid = 1;
+		return 0;
+	      }
+	}
+
+      if (multiplier != 0 && n * multiplier / multiplier != n)
+	{
+	  *invalid = 1;
+	  return 0;
+	}
+
+      n *= multiplier;
+    }
+}
+
+static void
+scanargs (int argc, char **argv)
+{
+  int i;
+  int c;
+
+  while ((c = getopt_long (argc, argv, "", long_options, NULL)) != -1)
+    {
+      switch (c)
+	{
+	case 0:
+	  break;
+
+	default:
+	  usage (1);
 	}
     }
-  else
-    {
-      while (records-- > 0)
-	{
-	  int nread;
 
-	  nread = safe_read (fdesc, buf, blocksize);
-	  if (nread < 0)
-	    {
-	      error (0, errno, "%s", file);
-	      quit (1);
-	    }
-	  /* POSIX doesn't say what to do when dd detects it has been
-	     asked to skip past EOF, so I assume it's non-fatal.
-	     FIXME: maybe give a warning.  */
-	  if (nread == 0)
-	    break;
+  for (i = optind; i < argc; i++)
+    {
+      char *name, *val;
+
+      name = argv[i];
+      val = strchr (name, '=');
+      if (val == NULL)
+	{
+	  error (0, 0, _("unrecognized option `%s'"), name);
+	  usage (1);
 	}
+      *val++ = '\0';
+
+      if (STREQ (name, "if"))
+	input_file = val;
+      else if (STREQ (name, "of"))
+	output_file = val;
+      else if (STREQ (name, "conv"))
+	parse_conversion (val);
+      else
+	{
+	  int invalid = 0;
+	  uintmax_t n = parse_integer (val, &invalid);
+
+	  if (STREQ (name, "ibs"))
+	    {
+	      input_blocksize = n;
+	      invalid |= input_blocksize != n || input_blocksize == 0;
+	      conversions_mask |= C_TWOBUFS;
+	    }
+	  else if (STREQ (name, "obs"))
+	    {
+	      output_blocksize = n;
+	      invalid |= output_blocksize != n || output_blocksize == 0;
+	      conversions_mask |= C_TWOBUFS;
+	    }
+	  else if (STREQ (name, "bs"))
+	    {
+	      output_blocksize = input_blocksize = n;
+	      invalid |= output_blocksize != n || output_blocksize == 0;
+	    }
+	  else if (STREQ (name, "cbs"))
+	    {
+	      conversion_blocksize = n;
+	      invalid |= (conversion_blocksize != n
+			  || conversion_blocksize == 0);
+	    }
+	  else if (STREQ (name, "skip"))
+	    skip_records = n;
+	  else if (STREQ (name, "seek"))
+	    seek_record = n;
+	  else if (STREQ (name, "count"))
+	    max_records = n;
+	  else
+	    {
+	      error (0, 0, _("unrecognized option `%s=%s'"), name, val);
+	      usage (1);
+	    }
+
+	  if (invalid)
+	    error (1, 0, _("invalid number `%s'"), val);
+	}
+    }
+
+  /* If bs= was given, both `input_blocksize' and `output_blocksize' will
+     have been set to positive values.  If either has not been set,
+     bs= was not given, so make sure two buffers are used. */
+  if (input_blocksize == 0 || output_blocksize == 0)
+    conversions_mask |= C_TWOBUFS;
+  if (input_blocksize == 0)
+    input_blocksize = DEFAULT_BLOCKSIZE;
+  if (output_blocksize == 0)
+    output_blocksize = DEFAULT_BLOCKSIZE;
+  if (conversion_blocksize == 0)
+    conversions_mask &= ~(C_BLOCK | C_UNBLOCK);
+}
+
+/* Fix up translation table. */
+
+static void
+apply_translations (void)
+{
+  int i;
+
+#define MX(a) (bit_count (conversions_mask & (a)))
+  if ((MX (C_ASCII | C_EBCDIC | C_IBM) > 1)
+      || (MX (C_BLOCK | C_UNBLOCK) > 1)
+      || (MX (C_LCASE | C_UCASE) > 1)
+      || (MX (C_UNBLOCK | C_SYNC) > 1))
+    {
+      error (1, 0, _("\
+only one conv in {ascii,ebcdic,ibm}, {lcase,ucase}, {block,unblock}, {unblock,sync}"));
+    }
+#undef MX
+
+  if (conversions_mask & C_ASCII)
+    translate_charset (ebcdic_to_ascii);
+
+  if (conversions_mask & C_UCASE)
+    {
+      for (i = 0; i < 256; i++)
+	if (ISLOWER (trans_table[i]))
+	  trans_table[i] = toupper (trans_table[i]);
+      translation_needed = 1;
+    }
+  else if (conversions_mask & C_LCASE)
+    {
+      for (i = 0; i < 256; i++)
+	if (ISUPPER (trans_table[i]))
+	  trans_table[i] = tolower (trans_table[i]);
+      translation_needed = 1;
+    }
+
+  if (conversions_mask & C_EBCDIC)
+    {
+      translate_charset (ascii_to_ebcdic);
+      newline_character = ascii_to_ebcdic['\n'];
+      space_character = ascii_to_ebcdic[' '];
+    }
+  else if (conversions_mask & C_IBM)
+    {
+      translate_charset (ascii_to_ibm);
+      newline_character = ascii_to_ibm['\n'];
+      space_character = ascii_to_ibm[' '];
     }
 }
 
@@ -517,18 +789,149 @@ swab_buffer (unsigned char *buf, int *nread)
   return ++bufstart;
 }
 
-/* Output buffer. */
-static unsigned char *obuf;
+/* Throw away RECORDS blocks of BLOCKSIZE bytes on file descriptor FDESC,
+   which is open with read permission for FILE.  Store up to BLOCKSIZE
+   bytes of the data at a time in BUF, if necessary. */
 
-/* Current index into `obuf'. */
-static int oc = 0;
+static void
+skip (int fdesc, char *file, uintmax_t records, size_t blocksize,
+      unsigned char *buf)
+{
+  struct stat stats;
+  off_t o;
 
-/* Index into current line, for `conv=block' and `conv=unblock'.  */
-static int col = 0;
+  /* Use fstat instead of checking for errno == ESPIPE because
+     lseek doesn't work on some special files but doesn't return an
+     error, either. */
+  /* FIXME: can this really happen?  What system?  */
+  if (fstat (fdesc, &stats))
+    {
+      error (0, errno, "%s", file);
+      quit (1);
+    }
+
+  /* Try lseek and if an error indicates it was an inappropriate
+     operation, fall back on using read.  */
+  o = records * blocksize;
+  if (o / blocksize != records
+      || lseek (fdesc, o, SEEK_SET) == -1)
+    {
+      while (records-- > 0)
+	{
+	  int nread;
+
+	  nread = safe_read (fdesc, buf, blocksize);
+	  if (nread < 0)
+	    {
+	      error (0, errno, "%s", file);
+	      quit (1);
+	    }
+	  /* POSIX doesn't say what to do when dd detects it has been
+	     asked to skip past EOF, so I assume it's non-fatal.
+	     FIXME: maybe give a warning.  */
+	  if (nread == 0)
+	    break;
+	}
+    }
+}
+
+/* Copy NREAD bytes of BUF, with no conversions.  */
+
+static void
+copy_simple (unsigned char *buf, int nread)
+{
+  int nfree;			/* Number of unused bytes in `obuf'.  */
+  unsigned char *start = buf; /* First uncopied char in BUF.  */
+
+  do
+    {
+      nfree = output_blocksize - oc;
+      if (nfree > nread)
+	nfree = nread;
+
+      memcpy ((char *) (obuf + oc), (char *) start, nfree);
+
+      nread -= nfree;		/* Update the number of bytes left to copy. */
+      start += nfree;
+      oc += nfree;
+      if (oc >= output_blocksize)
+	write_output ();
+    }
+  while (nread > 0);
+}
+
+/* Copy NREAD bytes of BUF, doing conv=block
+   (pad newline-terminated records to `conversion_blocksize',
+   replacing the newline with trailing spaces).  */
+
+static void
+copy_with_block (unsigned char *buf, int nread)
+{
+  register int i;
+
+  for (i = nread; i; i--, buf++)
+    {
+      if (*buf == newline_character)
+	{
+	  if (col < conversion_blocksize)
+	    {
+	      size_t j;
+	      for (j = col; j < conversion_blocksize; j++)
+		output_char (space_character);
+	    }
+	  col = 0;
+	}
+      else
+	{
+	  if (col == conversion_blocksize)
+	    r_truncate++;
+	  else if (col < conversion_blocksize)
+	    output_char (*buf);
+	  col++;
+	}
+    }
+}
+
+/* Copy NREAD bytes of BUF, doing conv=unblock
+   (replace trailing spaces in `conversion_blocksize'-sized records
+   with a newline).  */
+
+static void
+copy_with_unblock (unsigned char *buf, int nread)
+{
+  register int i;
+  register unsigned char c;
+  static int pending_spaces = 0;
+
+  for (i = 0; i < nread; i++)
+    {
+      c = buf[i];
+
+      if (col++ >= conversion_blocksize)
+	{
+	  col = pending_spaces = 0; /* Wipe out any pending spaces.  */
+	  i--;			/* Push the char back; get it later. */
+	  output_char (newline_character);
+	}
+      else if (c == space_character)
+	pending_spaces++;
+      else
+	{
+	  /* `c' is the character after a run of spaces that were not
+	     at the end of the conversion buffer.  Output them.  */
+	  while (pending_spaces)
+	    {
+	      output_char (space_character);
+	      --pending_spaces;
+	    }
+	  output_char (c);
+	}
+    }
+}
 
 /* The main loop.  */
 
-static void
+static int
 copy (void)
 {
   unsigned char *ibuf, *bufstart; /* Input buffer. */
@@ -548,18 +951,27 @@ copy (void)
   else
     obuf = ibuf;
 
-  if (skip_records > 0)
+  if (skip_records != 0)
     skip (input_fd, input_file, skip_records, input_blocksize, ibuf);
 
-  if (seek_record > 0)
-    skip (output_fd, output_file, seek_record, output_blocksize, obuf);
+  if (seek_record != 0)
+    {
+      /* FIXME: this loses for
+	 % ./dd if=dd seek=1 |:
+	 ./dd: a1 standard output: Bad file number
+	 0+0 records in
+	 0+0 records out
+	 */
+
+	 skip (output_fd, output_file, seek_record, output_blocksize, obuf);
+    }
 
   if (max_records == 0)
     quit (exit_status);
 
   while (1)
     {
-      if (max_records >= 0 && r_partial + r_full >= max_records)
+      if (r_partial + r_full >= max_records)
 	break;
 
       /* Zero the buffer before reading, so that if we get a read error,
@@ -580,7 +992,7 @@ copy (void)
 	    {
 	      print_stats ();
 	      /* Seek past the bad block if possible. */
-	      lseek (input_fd, input_blocksize, SEEK_CUR);
+	      lseek (input_fd, (off_t) input_blocksize, SEEK_CUR);
 	      if (conversions_mask & C_SYNC)
 		/* Replace the missing input with null bytes and
 		   proceed normally.  */
@@ -658,12 +1070,9 @@ copy (void)
     {
       /* If the final input line didn't end with a '\n', pad
 	 the output block to `conversion_blocksize' chars.  */
-      int pending_spaces = max (0, conversion_blocksize - col);
-      while (pending_spaces)
-	{
-	  output_char (space_character);
-	  --pending_spaces;
-	}
+      size_t i;
+      for (i = col; i < conversion_blocksize; i++)
+	output_char (space_character);
     }
 
   if ((conversions_mask & C_UNBLOCK) && col == conversion_blocksize)
@@ -672,7 +1081,7 @@ copy (void)
     output_char (newline_character);
 
   /* Write out the last block. */
-  if (oc > 0)
+  if (oc != 0)
     {
       int nwritten = full_write (output_fd, obuf, oc);
       if (nwritten > 0)
@@ -688,437 +1097,84 @@ copy (void)
   if (obuf != ibuf)
     free (obuf);
 
+  return exit_status;
+}
+
+int
+main (int argc, char **argv)
+{
+  int i;
+  int exit_status;
+
+  program_name = argv[0];
+  setlocale (LC_ALL, "");
+  bindtextdomain (PACKAGE, LOCALEDIR);
+  textdomain (PACKAGE);
+
+  /* Initialize translation table to identity translation. */
+  for (i = 0; i < 256; i++)
+    trans_table[i] = i;
+
+  /* Decode arguments. */
+  scanargs (argc, argv);
+
+  if (show_version)
+    {
+      printf ("dd (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
+      exit (0);
+    }
+
+  if (show_help)
+    usage (0);
+
+  apply_translations ();
+
+  if (input_file != NULL)
+    {
+      input_fd = open (input_file, O_RDONLY);
+      if (input_fd < 0)
+	error (1, errno, "%s", input_file);
+    }
+  else
+    input_file = _("standard input");
+
+  if (input_fd == output_fd)
+    error (1, 0, _("%s is closed"), (input_fd == 0
+				     ? _("standard input")
+				     : _("standard output")));
+
+  if (output_file != NULL)
+    {
+      int omode = O_RDWR | O_CREAT;
+
+      if (seek_record == 0 && !(conversions_mask & C_NOTRUNC))
+	omode |= O_TRUNC;
+      output_fd = open (output_file, omode, 0666);
+      if (output_fd < 0)
+	error (1, errno, "%s", output_file);
+#if HAVE_FTRUNCATE
+      if (seek_record != 0 && !(conversions_mask & C_NOTRUNC))
+	{
+	  off_t o = seek_record * output_blocksize;
+	  if (o / output_blocksize != seek_record)
+	    error (1, 0, _("file offset out of range"));
+	  if (ftruncate (output_fd, o) < 0)
+	    error (0, errno, "%s", output_file);
+	}
+#endif
+    }
+  else
+    {
+      output_file = _("standard output");
+    }
+
+  install_handler (SIGINT, interrupt_handler);
+  install_handler (SIGQUIT, interrupt_handler);
+  install_handler (SIGPIPE, interrupt_handler);
+  install_handler (SIGINFO, siginfo_handler);
+
+  exit_status = copy ();
+
   quit (exit_status);
 }
-
-/* Copy NREAD bytes of BUF, with no conversions.  */
-
-static void
-copy_simple (unsigned char *buf, int nread)
-{
-  int nfree;			/* Number of unused bytes in `obuf'.  */
-  unsigned char *start = buf; /* First uncopied char in BUF.  */
-
-  do
-    {
-      nfree = output_blocksize - oc;
-      if (nfree > nread)
-	nfree = nread;
-
-      memcpy ((char *) (obuf + oc), (char *) start, nfree);
-
-      nread -= nfree;		/* Update the number of bytes left to copy. */
-      start += nfree;
-      oc += nfree;
-      if (oc >= output_blocksize)
-	write_output ();
-    }
-  while (nread > 0);
-}
-
-/* Copy NREAD bytes of BUF, doing conv=block
-   (pad newline-terminated records to `conversion_blocksize',
-   replacing the newline with trailing spaces).  */
-
-static void
-copy_with_block (unsigned char *buf, int nread)
-{
-  register int i;
-
-  for (i = nread; i; i--, buf++)
-    {
-      if (*buf == newline_character)
-	{
-	  int pending_spaces = max (0, conversion_blocksize - col);
-	  while (pending_spaces)
-	    {
-	      output_char (space_character);
-	      --pending_spaces;
-	    }
-	  col = 0;
-	}
-      else
-	{
-	  if (col == conversion_blocksize)
-	    r_truncate++;
-	  else if (col < conversion_blocksize)
-	    output_char (*buf);
-	  col++;
-	}
-    }
-}
-
-/* Copy NREAD bytes of BUF, doing conv=unblock
-   (replace trailing spaces in `conversion_blocksize'-sized records
-   with a newline).  */
-
-static void
-copy_with_unblock (unsigned char *buf, int nread)
-{
-  register int i;
-  register unsigned char c;
-  static int pending_spaces = 0;
-
-  for (i = 0; i < nread; i++)
-    {
-      c = buf[i];
-
-      if (col++ >= conversion_blocksize)
-	{
-	  col = pending_spaces = 0; /* Wipe out any pending spaces.  */
-	  i--;			/* Push the char back; get it later. */
-	  output_char (newline_character);
-	}
-      else if (c == space_character)
-	pending_spaces++;
-      else
-	{
-	  /* `c' is the character after a run of spaces that were not
-	     at the end of the conversion buffer.  Output them.  */
-	  while (pending_spaces)
-	    {
-	      output_char (space_character);
-	      --pending_spaces;
-	    }
-	  output_char (c);
-	}
-    }
-}
-
-/* Write, then empty, the output buffer `obuf'. */
-
-static void
-write_output (void)
-{
-  int nwritten = full_write (output_fd, obuf, output_blocksize);
-  if (nwritten != output_blocksize)
-    {
-      error (0, errno, "%s", output_file);
-      if (nwritten > 0)
-	w_partial++;
-      quit (1);
-    }
-  else
-    w_full++;
-  oc = 0;
-}
-
-static void
-scanargs (int argc, char **argv)
-{
-  int i, n;
-  int c;
-
-  while ((c = getopt_long (argc, argv, "", long_options, (int *) 0)) != EOF)
-    {
-      switch (c)
-	{
-	case 0:
-	  break;
-
-	default:
-	  usage (1);
-	}
-    }
-
-  for (i = optind; i < argc; i++)
-    {
-      char *name, *val;
-
-      name = argv[i];
-      val = strchr (name, '=');
-      if (val == NULL)
-	{
-	  error (0, 0, _("unrecognized option `%s'"), name);
-	  usage (1);
-	}
-      *val++ = '\0';
-
-      if (equal (name, "if"))
-	input_file = val;
-      else if (equal (name, "of"))
-	output_file = val;
-      else if (equal (name, "conv"))
-	parse_conversion (val);
-      else
-	{
-	  n = parse_integer (val);
-	  if (n < 0)
-	    error (1, 0, _("invalid number `%s'"), val);
-
-	  if (equal (name, "ibs"))
-	    {
-	      input_blocksize = n;
-	      conversions_mask |= C_TWOBUFS;
-	    }
-	  else if (equal (name, "obs"))
-	    {
-	      output_blocksize = n;
-	      conversions_mask |= C_TWOBUFS;
-	    }
-	  else if (equal (name, "bs"))
-	    output_blocksize = input_blocksize = n;
-	  else if (equal (name, "cbs"))
-	    conversion_blocksize = n;
-	  else if (equal (name, "skip"))
-	    skip_records = n;
-	  else if (equal (name, "seek"))
-	    seek_record = n;
-	  else if (equal (name, "count"))
-	    max_records = n;
-	  else
-	    {
-	      error (0, 0, _("unrecognized option `%s=%s'"), name, val);
-	      usage (1);
-	    }
-	}
-    }
-
-  /* If bs= was given, both `input_blocksize' and `output_blocksize' will
-     have been set to non-negative values.  If either has not been set,
-     bs= was not given, so make sure two buffers are used. */
-  if (input_blocksize == -1 || output_blocksize == -1)
-    conversions_mask |= C_TWOBUFS;
-  if (input_blocksize == -1)
-    input_blocksize = DEFAULT_BLOCKSIZE;
-  if (output_blocksize == -1)
-    output_blocksize = DEFAULT_BLOCKSIZE;
-  if (conversion_blocksize == 0)
-    conversions_mask &= ~(C_BLOCK | C_UNBLOCK);
-}
-
-/* Return the value of STR, interpreted as a non-negative decimal integer,
-   optionally multiplied by various values.
-   Return -1 if STR does not represent a number in this format. */
-
-/* FIXME: use xstrtou?l */
-
-static int
-parse_integer (char *str)
-{
-  register int n = 0;
-  register int temp;
-  register char *p = str;
-
-  while (ISDIGIT (*p))
-    {
-      n = n * 10 + *p - '0';
-      p++;
-    }
-loop:
-  switch (*p++)
-    {
-    case '\0':
-      return n;
-    case 'b':
-      n *= 512;
-      goto loop;
-    case 'c':
-      goto loop;
-    case 'k':
-      n *= 1024;
-      goto loop;
-    case 'w':
-      n *= 2;
-      goto loop;
-    case 'x':
-      temp = parse_integer (p);
-      if (temp == -1)
-	return -1;
-      n *= temp;
-      break;
-    default:
-      return -1;
-    }
-  return n;
-}
-
-/* Interpret one "conv=..." option. */
-
-static void
-parse_conversion (char *str)
-{
-  char *new;
-  int i;
-
-  do
-    {
-      new = strchr (str, ',');
-      if (new != NULL)
-	*new++ = '\0';
-      for (i = 0; conversions[i].convname != NULL; i++)
-	if (equal (conversions[i].convname, str))
-	  {
-	    conversions_mask |= conversions[i].conversion;
-	    break;
-	  }
-      if (conversions[i].convname == NULL)
-	{
-	  error (0, 0, _("%s: invalid conversion"), str);
-	  usage (1);
-	}
-      str = new;
-  } while (new != NULL);
-}
-
-/* Fix up translation table. */
-
-static void
-apply_translations (void)
-{
-  int i;
-
-#define MX(a) (bit_count (conversions_mask & (a)))
-  if ((MX (C_ASCII | C_EBCDIC | C_IBM) > 1)
-      || (MX (C_BLOCK | C_UNBLOCK) > 1)
-      || (MX (C_LCASE | C_UCASE) > 1)
-      || (MX (C_UNBLOCK | C_SYNC) > 1))
-    {
-      error (1, 0, _("\
-only one conv in {ascii,ebcdic,ibm}, {lcase,ucase}, {block,unblock}, {unblock,sync}"));
-    }
-#undef MX
-
-  if (conversions_mask & C_ASCII)
-    translate_charset (ebcdic_to_ascii);
-
-  if (conversions_mask & C_UCASE)
-    {
-      for (i = 0; i < 256; i++)
-	if (ISLOWER (trans_table[i]))
-	  trans_table[i] = toupper (trans_table[i]);
-      translation_needed = 1;
-    }
-  else if (conversions_mask & C_LCASE)
-    {
-      for (i = 0; i < 256; i++)
-	if (ISUPPER (trans_table[i]))
-	  trans_table[i] = tolower (trans_table[i]);
-      translation_needed = 1;
-    }
-
-  if (conversions_mask & C_EBCDIC)
-    {
-      translate_charset (ascii_to_ebcdic);
-      newline_character = ascii_to_ebcdic['\n'];
-      space_character = ascii_to_ebcdic[' '];
-    }
-  else if (conversions_mask & C_IBM)
-    {
-      translate_charset (ascii_to_ibm);
-      newline_character = ascii_to_ibm['\n'];
-      space_character = ascii_to_ibm[' '];
-    }
-}
-
-static void
-translate_charset (const unsigned char *new_trans)
-{
-  int i;
-
-  for (i = 0; i < 256; i++)
-    trans_table[i] = new_trans[trans_table[i]];
-  translation_needed = 1;
-}
-
-/* Return the number of 1 bits in `i'. */
-
-static int
-bit_count (register unsigned int i)
-{
-  register int set_bits;
-
-  for (set_bits = 0; i != 0; set_bits++)
-    i &= i - 1;
-  return set_bits;
-}
-
-static void
-print_stats (void)
-{
-  fprintf (stderr, _("%u+%u records in\n"), r_full, r_partial);
-  fprintf (stderr, _("%u+%u records out\n"), w_full, w_partial);
-  if (r_truncate > 0)
-    fprintf (stderr, "%u %s\n", r_truncate,
-	     (r_truncate == 1
-	      ? _("truncated record")
-	      : _("truncated records")));
-}
-
-static void
-cleanup (void)
-{
-  print_stats ();
-  if (close (input_fd) < 0)
-    error (1, errno, "%s", input_file);
-  if (close (output_fd) < 0)
-    error (1, errno, "%s", output_file);
-}
-
-static void
-quit (int code)
-{
-  cleanup ();
-  exit (code);
-}
-
-static RETSIGTYPE
-interrupt_handler (int sig)
-{
-#ifdef SA_INTERRUPT
-  struct sigaction sigact;
-
-  sigact.sa_handler = SIG_DFL;
-  sigemptyset (&sigact.sa_mask);
-  sigact.sa_flags = 0;
-  sigaction (sig, &sigact, NULL);
-#else				/* !SA_INTERRUPT */
-  signal (sig, SIG_DFL);
-#endif				/* SA_INTERRUPT */
-  cleanup ();
-  kill (getpid (), sig);
-}
-
-static void
-usage (int status)
-{
-  if (status != 0)
-    fprintf (stderr, _("Try `%s --help' for more information.\n"),
-	     program_name);
-  else
-    {
-      printf (_("Usage: %s [OPTION]...\n"), program_name);
-      printf (_("\
-Copy a file, converting and formatting according to the options.\n\
-\n\
-  bs=BYTES        force ibs=BYTES and obs=BYTES\n\
-  cbs=BYTES       convert BYTES bytes at a time\n\
-  conv=KEYWORDS   convert the file as per the comma separated keyword list\n\
-  count=BLOCKS    copy only BLOCKS input blocks\n\
-  ibs=BYTES       read BYTES bytes at a time\n\
-  if=FILE         read from FILE instead of stdin\n\
-  obs=BYTES       write BYTES bytes at a time\n\
-  of=FILE         write to FILE instead of stdout, don't truncate file\n\
-  seek=BLOCKS     skip BLOCKS obs-sized blocks at start of output\n\
-  skip=BLOCKS     skip BLOCKS ibs-sized blocks at start of input\n\
-      --help      display this help and exit\n\
-      --version   output version information and exit\n\
-\n\
-BYTES may be suffixed: by xM for multiplication by M, by c for x1,\n\
-by w for x2, by b for x512, by k for x1024.  Each KEYWORD may be:\n\
-\n\
-  ascii     from EBCDIC to ASCII\n\
-  ebcdic    from ASCII to EBCDIC\n\
-  ibm       from ASCII to alternated EBCDIC\n\
-  block     pad newline-terminated records with spaces to cbs-size\n\
-  unblock   replace trailing spaces in cbs-size records with newline\n\
-  lcase     change upper case to lower case\n\
-  ucase     change lower case to upper case\n\
-  swab      swap every pair of input bytes\n\
-  noerror   continue after read errors\n\
-  sync      pad every input block with NULs to ibs-size\n\
-"));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
-    }
-  exit (status);
-}
