@@ -1,5 +1,5 @@
 /* du -- summarize disk usage
-   Copyright (C) 88, 89, 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 88, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,37 +26,39 @@
 		arguments have been processed.  This can be used to find
 		out the disk usage of a directory, with some files excluded.
    -h		Print sizes in human readable format (1k 234M 2G, etc).
-   -k		Print sizes in kilobytes instead of 512 byte blocks
-		(the default required by POSIX).
-   -m		Print sizes in megabytes instead of 512 byte blocks
+   -H		Similar, but use powers of 1000 not 1024.
+   -k		Print sizes in kilobytes.
+   -m		Print sizes in megabytes.
    -b		Print sizes in bytes.
    -S		Count the size of each directory separately, not including
 		the sizes of subdirectories.
    -D		Dereference only symbolic links given on the command line.
    -L		Dereference all symbolic links.
+   --exclude=PAT Exclude files that match PAT.
+   -X FILE	Exclude files that match patterns taken from FILE.
 
    By tege@sics.se, Torbjorn Granlund,
    and djm@ai.mit.edu, David MacKenzie.
-   Variable blocks added by lm@sgi.com.
+   Variable blocks added by lm@sgi.com and eggert@twinsun.com.
 */
 
-#ifdef _AIX
- #pragma alloca
-#endif
-
 #include <config.h>
+#if HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
 #include <stdio.h>
 #include <getopt.h>
 #include <sys/types.h>
 #include <assert.h>
 
+#include "exclude.h"
 #include "system.h"
 #include "save-cwd.h"
+#include "closeout.h"
 #include "error.h"
-
-#undef	convert_blocks
-#define	convert_blocks(b, size) (size == size_kilobytes ? ((b) + 1) / 2 : \
-    size == size_megabytes ? ((b) + 1024) / 2048 : (b))
+#include "human.h"
+#include "xstrtol.h"
+#include "savedir.h"
 
 /* Initial number of entries in each hash table entry's table of inodes.  */
 #define INITIAL_HASH_MODULE 100
@@ -67,11 +69,6 @@
 /* Initial size to allocate for `path'.  */
 #define INITIAL_PATH_SIZE 100
 
-/* The maximum length of a human-readable string.  Be pessimistic
-   and assume `int' is 64-bits wide.  Converting 2^63 - 1 gives the
-   11-character string, 8589934592G.  */
-#define LONGEST_HUMAN_READABLE 11
-
 /* Hash structure for inode and device numbers.  The separate entry
    structure makes it easier to rehash "in place".  */
 
@@ -96,38 +93,33 @@ struct htab
 
 /* Structure for dynamically resizable strings. */
 
-typedef struct
+struct String
 {
   unsigned alloc;		/* Size of allocation for the text.  */
   unsigned length;		/* Length of the text currently.  */
   char *text;			/* Pointer to the text.  */
-} *string, stringstruct;
+};
+typedef struct String String;
 
 int stat ();
 int lstat ();
 
-char *savedir ();
-char *xmalloc ();
-char *xrealloc ();
-
-static int hash_insert __P ((ino_t ino, dev_t dev));
-static int hash_insert2 __P ((struct htab *htab, ino_t ino, dev_t dev));
-static long count_entry __P ((char *ent, int top, dev_t last_dev));
-static void du_files __P ((char **files));
-static void hash_init __P ((unsigned int modulus,
-			    unsigned int entry_tab_size));
-static void hash_reset __P ((void));
-static void str_concatc __P ((string s1, char *cstr));
-static void str_copyc __P ((string s1, char *cstr));
-static void str_init __P ((string *s1, unsigned int size));
-static void str_trunc __P ((string s1, unsigned int length));
+static int hash_insert PARAMS ((ino_t ino, dev_t dev));
+static int hash_insert2 PARAMS ((struct htab *_htab, ino_t ino, dev_t dev));
+static uintmax_t count_entry PARAMS ((const char *ent, int top, dev_t last_dev,
+				      int depth));
+static void du_files PARAMS ((char **files));
+static void hash_init PARAMS ((unsigned int modulus,
+			       unsigned int entry_tab_size));
+static void hash_reset PARAMS ((void));
+static void str_concatc PARAMS ((String *s1, char *cstr));
+static void str_copyc PARAMS ((String *s1, char *cstr));
+static void str_init PARAMS ((String **s1, unsigned int size));
+static void str_trunc PARAMS ((String *s1, unsigned int length));
 
 /* Name under which this program was invoked.  */
 char *program_name;
 
-/* If nonzero, display only a total for each argument. */
-static int opt_summarize_only = 0;
-
 /* If nonzero, display counts for all files, not just directories. */
 static int opt_all = 0;
 
@@ -146,22 +138,17 @@ static int opt_separate_dirs = 0;
 /* If nonzero, dereference symlinks that are command line arguments. */
 static int opt_dereference_arguments = 0;
 
-enum output_size
-{
-  size_blocks,			/* 512-byte blocks. */
-  size_kilobytes,		/* 1K blocks. */
-  size_megabytes,		/* 1024K blocks. */
-  size_bytes			/* 1-byte blocks. */
-};
+/* Show the total for each directory (and file if --all) that is at
+   most MAX_DEPTH levels down from the root of the hierarchy.  The root
+   is at level 0, so `du --max-depth=0' is equivalent to `du -s'.  */
+static int max_depth = INT_MAX;
 
-/* human style output */
-static int opt_human_readable;
-
-/* The units to count in. */
-static enum output_size output_size;
+/* If positive, the units to use when printing sizes;
+   if negative, the human-readable base.  */
+static int output_block_size;
 
 /* Accumulated path for file or directory being processed.  */
-static string path;
+static String *path;
 
 /* Pointer to hash structure, used by the hash routines.  */
 static struct htab *htab;
@@ -182,23 +169,31 @@ static int show_help;
 /* If nonzero, print the version on standard output and exit.  */
 static int show_version;
 
-/* Grand total size of all args. */
-static long tot_size = 0L;
+/* File name patterns to exclude.  */
+static struct exclude *exclude;
+
+/* Grand total size of all args, in units of ST_NBLOCKSIZE-byte blocks. */
+static uintmax_t tot_size = 0;
 
 static struct option const long_options[] =
 {
-  {"all", no_argument, &opt_all, 1},
+  {"all", no_argument, NULL, 'a'},
+  {"block-size", required_argument, 0, CHAR_MAX + 2},
   {"bytes", no_argument, NULL, 'b'},
-  {"count-links", no_argument, &opt_count_all, 1},
+  {"count-links", no_argument, NULL, 'l'},
   {"dereference", no_argument, NULL, 'L'},
-  {"dereference-args", no_argument, &opt_dereference_arguments, 1},
+  {"dereference-args", no_argument, NULL, 'D'},
+  {"exclude", required_argument, 0, CHAR_MAX + 1},
+  {"exclude-from", required_argument, 0, 'X'},
   {"human-readable", no_argument, NULL, 'h'},
+  {"si", no_argument, 0, 'H'},
   {"kilobytes", no_argument, NULL, 'k'},
+  {"max-depth", required_argument, NULL, CHAR_MAX + 3},
   {"megabytes", no_argument, NULL, 'm'},
-  {"one-file-system", no_argument, &opt_one_file_system, 1},
-  {"separate-dirs", no_argument, &opt_separate_dirs, 1},
-  {"summarize", no_argument, &opt_summarize_only, 1},
-  {"total", no_argument, &opt_combined_arguments, 1},
+  {"one-file-system", no_argument, NULL, 'x'},
+  {"separate-dirs", no_argument, NULL, 'S'},
+  {"summarize", no_argument, NULL, 's'},
+  {"total", no_argument, NULL, 'c'},
   {"help", no_argument, &show_help, 1},
   {"version", no_argument, &show_version, 1},
   {NULL, 0, NULL, 0}
@@ -221,21 +216,30 @@ usage (int status, char *reason)
 Summarize disk usage of each FILE, recursively for directories.\n\
 \n\
   -a, --all             write counts for all files, not just directories\n\
+      --block-size=SIZE use SIZE-byte blocks\n\
   -b, --bytes           print size in bytes\n\
   -c, --total           produce a grand total\n\
   -D, --dereference-args  dereference PATHs when symbolic link\n\
   -h, --human-readable  print sizes in human readable format (e.g., 1K 234M 2G)\n\
-  -k, --kilobytes       use 1024-byte blocks, not 512 despite POSIXLY_CORRECT\n\
+  -H, --si              likewise, but use powers of 1000 not 1024\n\
+  -k, --kilobytes       like --block-size=1024\n\
   -l, --count-links     count sizes many times if hard linked\n\
   -L, --dereference     dereference all symbolic links\n\
-  -m, --megabytes       use 1024K-byte blocks, not 512 despite POSIXLY_CORRECT\n\
+  -m, --megabytes       like --block-size=1048576\n\
   -S, --separate-dirs   do not include size of subdirectories\n\
   -s, --summarize       display only a total for each argument\n\
   -x, --one-file-system  skip directories on different filesystems\n\
+  -X FILE, --exclude-from=FILE  Exclude files that match any pattern in FILE.\n\
+      --exclude=PAT     Exclude files that match PAT.\n\
+      --max-depth=N     print the total for a directory (or file, with --all)\n\
+                          only if it is N or fewer levels below the command\n\
+                          line argument;  --max-depth=0 is the same as\n\
+                          --summarize\n\
       --help            display this help and exit\n\
       --version         output version information and exit\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -245,7 +249,10 @@ main (int argc, char **argv)
 {
   int c;
   char *cwd_only[2];
-  char *bs;
+  int max_depth_specified = 0;
+
+  /* If nonzero, display only a total for each argument. */
+  int opt_summarize_only = 0;
 
   cwd_only[0] = ".";
   cwd_only[1] = NULL;
@@ -255,23 +262,15 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  exclude = new_exclude ();
   xstat = lstat;
 
-  if (getenv ("POSIXLY_CORRECT"))
-    output_size = size_blocks;
-  else if ((bs = getenv ("BLOCKSIZE"))
-	   && strncmp (bs, "HUMAN", sizeof ("HUMAN") - 1) == 0)
-    {
-      opt_human_readable = 1;
-      output_size = size_bytes;
-    }
-  else
-    output_size = size_kilobytes;
+  human_block_size (getenv ("DU_BLOCK_SIZE"), 0, &output_block_size);
 
-  while ((c = getopt_long (argc, argv, "abchklmsxDLS", long_options,
-			   (int *) 0))
-	 != EOF)
+  while ((c = getopt_long (argc, argv, "abchHklmsxDLSX:", long_options, NULL))
+	 != -1)
     {
+      long int tmp_long;
       switch (c)
 	{
 	case 0:			/* Long option. */
@@ -282,8 +281,7 @@ main (int argc, char **argv)
 	  break;
 
 	case 'b':
-	  output_size = size_bytes;
-	  opt_human_readable = 0;
+	  output_block_size = 1;
 	  break;
 
 	case 'c':
@@ -291,18 +289,28 @@ main (int argc, char **argv)
 	  break;
 
 	case 'h':
-	  output_size = size_bytes;
-	  opt_human_readable = 1;
+	  output_block_size = -1024;
+	  break;
+
+	case 'H':
+	  output_block_size = -1000;
 	  break;
 
 	case 'k':
-	  output_size = size_kilobytes;
-	  opt_human_readable = 0;
+	  output_block_size = 1024;
 	  break;
 
+	case CHAR_MAX + 3:		/* --max-depth=N */
+	  if (xstrtol (optarg, NULL, 0, &tmp_long, NULL) != LONGINT_OK
+	      || tmp_long < 0 || tmp_long > INT_MAX)
+	    error (1, 0, _("invalid maximum depth `%s'"), optarg);
+
+	  max_depth_specified = 1;
+	  max_depth = (int) tmp_long;
+ 	  break;
+
 	case 'm':
-	  output_size = size_megabytes;
-	  opt_human_readable = 0;
+	  output_block_size = 1024 * 1024;
 	  break;
 
 	case 'l':
@@ -329,14 +337,28 @@ main (int argc, char **argv)
 	  opt_separate_dirs = 1;
 	  break;
 
+	case 'X':
+	  if (add_exclude_file (exclude, optarg, '\n') != 0)
+	    error (1, errno, "%s", optarg);
+	  break;
+
+	case CHAR_MAX + 1:
+	  add_exclude (exclude, optarg);
+	  break;
+
+	case CHAR_MAX + 2:
+	  human_block_size (optarg, 1, &output_block_size);
+	  break;
+
 	default:
-	  usage (2, (char *) 0);
+	  usage (1, (char *) 0);
 	}
     }
 
   if (show_version)
     {
       printf ("du (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
@@ -344,7 +366,24 @@ main (int argc, char **argv)
     usage (0, NULL);
 
   if (opt_all && opt_summarize_only)
-    usage (2, _("cannot both summarize and show all entries"));
+    usage (1, _("cannot both summarize and show all entries"));
+
+  if (opt_summarize_only && max_depth_specified && max_depth == 0)
+    {
+      error (0, 0,
+	     _("warning: summarizing is the same as using --max-depth=0"));
+    }
+
+  if (opt_summarize_only && max_depth_specified && max_depth != 0)
+    {
+      error (0, 0,
+	     _("warning: summarizing conflicts with --max-depth=%d"),
+	       max_depth);
+      usage (1, NULL);
+    }
+
+  if (opt_summarize_only)
+    max_depth = 0;
 
   /* Initialize the hash structure for inode numbers.  */
   hash_init (INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);
@@ -353,63 +392,23 @@ main (int argc, char **argv)
 
   du_files (optind == argc ? cwd_only : argv + optind);
 
+  close_stdout ();
   exit (exit_status);
 }
 
-/* Convert N_BYTES to a more readable string than %d would.
-   Most people visually process strings of 3-4 digits effectively,
-   but longer strings of digits are more prone to misinterpretation.
-   Hence, converting to an abbreviated form usually improves readability.
-   Use a suffix indicating multiples of 1024 (K), 1024*1024 (M), and
-   1024*1024*1024 (G).  For example, 8500 would be converted to 8.3K,
-   133456345 to 127M, 56990456345 to 53G, and so on.  Numbers smaller
-   than 1024 aren't modified.  */
+/* Print N_BLOCKS followed by STRING on a line.  NBLOCKS is the number of
+   ST_NBLOCKSIZE-byte blocks; convert it to OUTPUT_BLOCK_SIZE units before
+   printing.  If OUTPUT_BLOCK_SIZE is negative, use a human readable
+   notation instead.  */
 
-static char *
-human_readable (int n_bytes, char *buf, int buf_len)
+static void
+print_size (uintmax_t n_blocks, const char *string)
 {
-  const char *suffix;
-  double amt;
-  char *p;
-
-  assert (buf_len > LONGEST_HUMAN_READABLE);
-
-  p = buf;
-  amt = n_bytes;
-
-  if (amt >= 1024 * 1024 * 1024)
-    {
-      amt /= (1024 * 1024 * 1024);
-      suffix = "G";
-    }
-  else if (amt >= 1024 * 1024)
-    {
-      amt /= (1024 * 1024);
-      suffix = "M";
-    }
-  else if (amt >= 1024)
-    {
-      amt /= 1024;
-      suffix = "K";
-    }
-  else
-    {
-      suffix = "";
-    }
-
-  if (amt >= 10)
-    {
-      sprintf (p, "%.0f%s", amt, suffix);
-    }
-  else if (amt == 0)
-    {
-      strcpy (p, "0");
-    }
-  else
-    {
-      sprintf (p, "%.1f%s", amt, suffix);
-    }
-  return (p);
+  char buf[LONGEST_HUMAN_READABLE + 1];
+  printf ("%s\t%s\n",
+	  human_readable (n_blocks, buf, ST_NBLOCKSIZE, output_block_size),
+	  string);
+  fflush (stdout);
 }
 
 /* Recursively print the sizes of the directories (and, if selected, files)
@@ -456,7 +455,7 @@ du_files (char **files)
       if (!opt_combined_arguments)
 	hash_reset ();
 
-      count_entry (arg, 1, 0);
+      count_entry (arg, 1, 0, 0);
 
       /* chdir if `count_entry' has changed the working directory.  */
       if (stat (".", &stat_buf))
@@ -469,33 +468,22 @@ du_files (char **files)
     }
 
   if (opt_combined_arguments)
-    {
-      if (opt_human_readable)
-	{
-	  char buf[LONGEST_HUMAN_READABLE + 1];
-	  printf("%s\ttotal\n", human_readable (tot_size, buf,
-						LONGEST_HUMAN_READABLE + 1));
-	}
-      else
-	{
-	  printf (_("%ld\ttotal\n"), output_size == size_bytes ? tot_size
-		  : convert_blocks (tot_size, output_size == size_kilobytes));
-	}
-      fflush (stdout);
-    }
+    print_size (tot_size, _("total"));
 
   free_cwd (&cwd);
 }
 
-/* Print (if appropriate) and return the size
-   (in units determined by `output_size') of file or directory ENT.
-   TOP is one for external calls, zero for recursive calls.
-   LAST_DEV is the device that the parent directory of ENT is on.  */
+/* Print (if appropriate) the size (in units determined by `output_block_size')
+   of file or directory ENT. Return the size of ENT in units of 512-byte
+   blocks.  TOP is one for external calls, zero for recursive calls.
+   LAST_DEV is the device that the parent directory of ENT is on.
+   DEPTH is the number of levels (in hierarchy) down from a command
+   line argument.  Don't print if DEPTH > max_depth.  */
 
-static long
-count_entry (char *ent, int top, dev_t last_dev)
+static uintmax_t
+count_entry (const char *ent, int top, dev_t last_dev, int depth)
 {
-  long size;
+  uintmax_t size;
 
   if (((top && opt_dereference_arguments)
        ? stat (ent, &stat_buf)
@@ -511,11 +499,7 @@ count_entry (char *ent, int top, dev_t last_dev)
       && hash_insert (stat_buf.st_ino, stat_buf.st_dev))
     return 0;			/* Have counted this already.  */
 
-  if (output_size == size_bytes)
-    size = stat_buf.st_size;
-  else
-    size = ST_NBLOCKS (stat_buf);
-
+  size = ST_NBLOCKS (stat_buf);
   tot_size += size;
 
   if (S_ISDIR (stat_buf.st_mode))
@@ -533,8 +517,8 @@ count_entry (char *ent, int top, dev_t last_dev)
       if (opt_one_file_system && !top && last_dev != dir_dev)
 	return 0;		/* Don't enter a new file system.  */
 
-#ifndef S_ISDIR
-# define S_ISDIR(s) 0
+#ifndef S_ISLNK
+# define S_ISLNK(s) 0
 #endif
       /* If we're dereferencing symlinks and we're about to chdir through
 	 a symlink, remember the current directory so we can return to it
@@ -542,9 +526,8 @@ count_entry (char *ent, int top, dev_t last_dev)
       through_symlink = (xstat == stat
 			 && lstat (ent, &e_buf) == 0
 			 && S_ISLNK (e_buf.st_mode));
-      if (through_symlink)
-	if (save_cwd (&cwd))
-	  exit (1);
+      if (through_symlink && save_cwd (&cwd))
+	exit (1);
 
       if (chdir (ent) < 0)
 	{
@@ -554,7 +537,7 @@ count_entry (char *ent, int top, dev_t last_dev)
 	}
 
       errno = 0;
-      name_space = savedir (".", stat_buf.st_size);
+      name_space = savedir (".", (unsigned int) stat_buf.st_size);
       if (name_space == NULL)
 	{
 	  if (errno)
@@ -581,16 +564,16 @@ count_entry (char *ent, int top, dev_t last_dev)
       str_concatc (path, "/");
       pathlen = path->length;
 
-      namep = name_space;
-      while (*namep != 0)
+      for (namep = name_space; *namep; namep += strlen (namep) + 1)
 	{
-	  str_concatc (path, namep);
-
-	  size += count_entry (namep, 0, dir_dev);
-
-	  str_trunc (path, pathlen);
-	  namep += strlen (namep) + 1;
+	  if (!excluded_filename (exclude, namep))
+	    {
+	      str_concatc (path, namep);
+	      size += count_entry (namep, 0, dir_dev, depth + 1);
+	      str_trunc (path, pathlen);
+	    }
 	}
+
       free (name_space);
       if (through_symlink)
 	{
@@ -598,51 +581,22 @@ count_entry (char *ent, int top, dev_t last_dev)
 	  free_cwd (&cwd);
 	}
       else if (chdir ("..") < 0)
-        error (1, errno,
-	       _("cannot change to `..' from directory %s"), path->text);
+	{
+	  error (1, errno,
+		 _("cannot change to `..' from directory %s"), path->text);
+	}
 
       str_trunc (path, pathlen - 1); /* Remove the "/" we added.  */
-      if (!opt_summarize_only || top)
-	{
-	  if (opt_human_readable)
-	    {
-	      char buf[LONGEST_HUMAN_READABLE + 1];
-	      printf("%s\t%s\n",
-		     human_readable (size, buf, LONGEST_HUMAN_READABLE + 1),
-		     path->length > 0 ? path->text : "/");
-	    }
-	  else
-	    {
-	      printf ("%ld\t%s\n", (output_size == size_bytes
-				    ? size
-				    : convert_blocks (size, output_size)),
-		      path->length > 0 ? path->text : "/");
-	    }
-	  fflush (stdout);
-	}
+      if (depth <= max_depth || top)
+	print_size (size, path->length > 0 ? path->text : "/");
       return opt_separate_dirs ? 0 : size;
     }
-  else if (opt_all || top)
+  else if ((opt_all && depth <= max_depth) || top)
     {
       /* FIXME: make this an option.  */
       int print_only_dir_size = 0;
       if (!print_only_dir_size)
-	{
-	  if (opt_human_readable)
-	    {
-	      char buf[LONGEST_HUMAN_READABLE + 1];
-	      printf("%s\t%s\n",
-		     human_readable (size, buf, LONGEST_HUMAN_READABLE + 1),
-		     path->length > 0 ? path->text : "/");
-	    }
-	  else
-	    {
-	      printf ("%ld\t%s\n", output_size == size_bytes ? size
-		      : convert_blocks (size, output_size),
-		      path->text);
-	    }
-	  fflush (stdout);
-	}
+	print_size (size, path->length > 0 ? path->text : "/");
     }
 
   return size;
@@ -723,7 +677,7 @@ hash_insert (ino_t ino, dev_t dev)
 
       htab_r->entry_tab = (struct entry *)
 	xrealloc ((char *) htab_r->entry_tab,
-		 sizeof (struct entry) * entry_tab_size);
+		  sizeof (struct entry) * entry_tab_size);
 
       /* Increase the size of htab again.  */
 
@@ -762,10 +716,10 @@ hash_insert (ino_t ino, dev_t dev)
    already existed.  */
 
 static int
-hash_insert2 (struct htab *htab, ino_t ino, dev_t dev)
+hash_insert2 (struct htab *ht, ino_t ino, dev_t dev)
 {
   struct entry **hp, *ep2, *ep;
-  hp = &htab->hash[ino % htab->modulus];
+  hp = &ht->hash[ino % ht->modulus];
   ep2 = *hp;
 
   /* Collision?  */
@@ -788,7 +742,7 @@ hash_insert2 (struct htab *htab, ino_t ino, dev_t dev)
 
     }
 
-  ep = *hp = &htab->entry_tab[htab->first_free_entry++];
+  ep = *hp = &ht->entry_tab[ht->first_free_entry++];
   ep->ino = ino;
   ep->dev = dev;
   ep->coll_link = ep2;		/* `ep2' is NULL if no collision.  */
@@ -796,14 +750,14 @@ hash_insert2 (struct htab *htab, ino_t ino, dev_t dev)
   return 0;
 }
 
-/* Initialize the struct string S1 for holding SIZE characters.  */
+/* Initialize string S1 to hold SIZE characters.  */
 
 static void
-str_init (string *s1, unsigned int size)
+str_init (String **s1, unsigned int size)
 {
-  string s;
+  String *s;
 
-  s = (string) xmalloc (sizeof (stringstruct));
+  s = (String *) xmalloc (sizeof (struct String));
   s->text = xmalloc (size + 1);
 
   s->alloc = size;
@@ -811,7 +765,7 @@ str_init (string *s1, unsigned int size)
 }
 
 static void
-ensure_space (string s, unsigned int size)
+ensure_space (String *s, unsigned int size)
 {
   if (s->alloc < size)
     {
@@ -823,7 +777,7 @@ ensure_space (string s, unsigned int size)
 /* Assign the null-terminated C-string CSTR to S1.  */
 
 static void
-str_copyc (string s1, char *cstr)
+str_copyc (String *s1, char *cstr)
 {
   unsigned l = strlen (cstr);
   ensure_space (s1, l);
@@ -832,7 +786,7 @@ str_copyc (string s1, char *cstr)
 }
 
 static void
-str_concatc (string s1, char *cstr)
+str_concatc (String *s1, char *cstr)
 {
   unsigned l1 = s1->length;
   unsigned l2 = strlen (cstr);
@@ -846,7 +800,7 @@ str_concatc (string s1, char *cstr)
 /* Truncate the string S1 to have length LENGTH.  */
 
 static void
-str_trunc (string s1, unsigned int length)
+str_trunc (String *s1, unsigned int length)
 {
   if (s1->length > length)
     {
