@@ -1,5 +1,5 @@
 /* install - copy files and set attributes
-   Copyright (C) 89, 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -42,6 +42,9 @@
    -s, --strip
 	Strip the symbol tables from installed files.
 
+   -p, --preserve-timestamps
+        Retain creation and modification timestamps when installing files.
+
    -d, --directory
 	Create a directory and its leading directories, if they
 	do not already exist.  Set the owner, group and mode
@@ -50,6 +53,11 @@
 	This is different from the SunOS 4.0 install, which gives
 	directories that it creates the default attributes.
 
+  -D
+	Like the -d option, but a file is installed, along with the directory.
+	Useful when installing into a new directory, and the install
+	process doesn't properly comprehend making directories.
+
    David MacKenzie <djm@gnu.ai.mit.edu> */
 
 #ifdef _AIX
@@ -67,8 +75,12 @@
 #include "backupfile.h"
 #include "modechange.h"
 #include "makepath.h"
+#include "closeout.h"
 #include "error.h"
 #include "xstrtol.h"
+#include "path-concat.h"
+#include "cp-hash.h"
+#include "copy.h"
 
 #if HAVE_SYS_WAIT_H
 # include <sys/wait.h>
@@ -78,10 +90,6 @@
 # include <values.h>
 #endif
 
-#ifndef BITSPERBYTE
-# define BITSPERBYTE 8
-#endif
-
 struct passwd *getpwnam ();
 struct group *getgrnam ();
 
@@ -90,14 +98,20 @@ uid_t getuid ();
 gid_t getgid ();
 #endif
 
-#ifndef HAVE_ENDGRENT
+#if ! HAVE_ENDGRENT
 # define endgrent() ((void) 0)
 #endif
 
-#ifndef HAVE_ENDPWENT
+#if ! HAVE_ENDPWENT
 # define endpwent() ((void) 0)
 #endif
 
+/* Initial number of entries in each hash table entry's table of inodes.  */
+#define INITIAL_HASH_MODULE 100
+
+/* Initial number of entries in the inode hash table.  */
+#define INITIAL_ENTRY_TAB_SIZE 70
+
 /* True if C is an ASCII octal digit. */
 #define isodigit(c) ((c) >= '0' && c <= '7')
 
@@ -105,34 +119,41 @@ gid_t getgid ();
 #define READ_SIZE (32 * 1024)
 
 #ifndef UID_T_MAX
-# define UID_T_MAX ((uid_t)(~((unsigned long)1 << ((sizeof (uid_t) \
-						    * BITSPERBYTE - 1)))))
+# define UID_T_MAX TYPE_MAXIMUM (uid_t)
 #endif
 
 #ifndef GID_T_MAX
-# define GID_T_MAX ((gid_t)(~((unsigned long)1 << ((sizeof (gid_t) \
-						    * BITSPERBYTE - 1)))))
+# define GID_T_MAX TYPE_MAXIMUM (gid_t)
 #endif
 
-char *basename ();
-char *stpcpy ();
-char *xmalloc ();
-int safe_read ();
+char *base_name ();
+char *dirname ();
 int full_write ();
 int isdir ();
 enum backup_type get_version ();
 
-static int change_attributes __P ((char *path, int no_need_to_chown));
-static int copy_file __P ((char *from, char *to, int *to_created));
-static int install_file_in_dir __P ((char *from, char *to_dir));
-static int install_file_in_file __P ((char *from, char *to));
-static void get_ids __P ((void));
-static void strip __P ((char *path));
-static void usage __P ((int status));
+int stat ();
+
+static int change_timestamps PARAMS ((const char *from, const char *to));
+static int change_attributes PARAMS ((const char *path));
+static int copy_file PARAMS ((const char *from, const char *to,
+			      const struct cp_options *x));
+static int install_file_to_path PARAMS ((const char *from, const char *to,
+					 const struct cp_options *x));
+static int install_file_in_dir PARAMS ((const char *from, const char *to_dir,
+					const struct cp_options *x));
+static int install_file_in_file PARAMS ((const char *from, const char *to,
+					 const struct cp_options *x));
+static void get_ids PARAMS ((void));
+static void strip PARAMS ((const char *path));
+static void usage PARAMS ((int status));
 
 /* The name this program was run with, for error messages. */
 char *program_name;
 
+/* FIXME: document */
+enum backup_type backup_type;
+
 /* The user name that will own the files, or NULL to make the owner
    the current user ID. */
 static char *owner_name;
@@ -149,7 +170,7 @@ static gid_t group_id;
 
 /* The permissions to which the files will be set.  The umask has
    no effect. */
-static int mode;
+static mode_t mode = 0755;
 
 /* If nonzero, strip executable files after copying them. */
 static int strip_files;
@@ -170,13 +191,51 @@ static struct option const long_options[] =
   {"group", required_argument, NULL, 'g'},
   {"mode", required_argument, NULL, 'm'},
   {"owner", required_argument, NULL, 'o'},
+  {"preserve-timestamps", no_argument, NULL, 'p'},
   {"backup", no_argument, NULL, 'b'},
+  {"suffix", required_argument, NULL, 'S'},
   {"version-control", required_argument, NULL, 'V'},
+  {"verbose", no_argument, NULL, 'v'},
   {"help", no_argument, &show_help, 1},
   {"version", no_argument, &show_version, 1},
   {NULL, 0, NULL, 0}
 };
 
+static void
+cp_option_init (struct cp_options *x)
+{
+  x->copy_as_regular = 1;
+  x->dereference = 1;
+  x->force = 1;
+
+  /* If unlink fails, try to proceed anyway.  */
+  x->failed_unlink_is_fatal = 0;
+
+  x->hard_link = 0;
+  x->interactive = 0;
+  x->move_mode = 0;
+  x->myeuid = geteuid ();
+  x->one_file_system = 0;
+  x->preserve_owner_and_group = 0;
+  x->preserve_chmod_bits = 0;
+  x->preserve_timestamps = 0;
+  x->require_preserve = 0;
+  x->recursive = 0;
+  x->sparse_mode = SPARSE_AUTO;
+  x->symbolic_link = 0;
+
+  /* Create destination files initially writable so we can run strip on them.
+     Although GNU strip works fine on read-only files, some others
+     would fail.  */
+  x->set_mode = 1;
+  x->mode = 0600;
+
+  x->umask_kill = 0;
+  x->update = 0;
+  x->verbose = 0;
+  x->xstat = stat;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -185,26 +244,31 @@ main (int argc, char **argv)
   char *symbolic_mode = NULL;
   int make_backups = 0;
   char *version;
+  int mkdir_and_install = 0;
+  struct cp_options x;
+  int n_files;
+  char **file;
 
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  cp_option_init (&x);
+
   owner_name = NULL;
   group_name = NULL;
-  mode = 0755;
   strip_files = 0;
   dir_arg = 0;
   umask (0);
 
-   version = getenv ("SIMPLE_BACKUP_SUFFIX");
-   if (version)
-      simple_backup_suffix = version;
-   version = getenv ("VERSION_CONTROL");
+  version = getenv ("SIMPLE_BACKUP_SUFFIX");
+  if (version)
+    simple_backup_suffix = version;
+  version = getenv ("VERSION_CONTROL");
 
-  while ((optc = getopt_long (argc, argv, "bcsdg:m:o:V:S:", long_options,
-			      (int *) 0)) != EOF)
+  while ((optc = getopt_long (argc, argv, "bcsDdg:m:o:pvV:S:", long_options,
+			      NULL)) != -1)
     {
       switch (optc)
 	{
@@ -221,6 +285,12 @@ main (int argc, char **argv)
 	case 'd':
 	  dir_arg = 1;
 	  break;
+	case 'D':
+	  mkdir_and_install = 1;
+	  break;
+	case 'v':
+	  x.verbose = 1;
+	  break;
 	case 'g':
 	  group_name = optarg;
 	  break;
@@ -230,6 +300,9 @@ main (int argc, char **argv)
 	case 'o':
 	  owner_name = optarg;
 	  break;
+	case 'p':
+	  x.preserve_timestamps = 1;
+	  break;
 	case 'S':
 	  simple_backup_suffix = optarg;
 	  break;
@@ -244,6 +317,7 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("install (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
@@ -255,10 +329,12 @@ main (int argc, char **argv)
     error (1, 0,
 	   _("the strip option may not be used when installing a directory"));
 
-  if (make_backups)
-    backup_type = get_version (version);
+  x.backup_type = (make_backups ? get_version (version) : none);
 
-  if (optind == argc || (optind == argc - 1 && !dir_arg))
+  n_files = argc - optind;
+  file = argv + optind;
+
+  if (n_files == 0 || (n_files == 1 && !dir_arg))
     {
       error (0, 0, _("too few arguments"));
       usage (1);
@@ -278,53 +354,107 @@ main (int argc, char **argv)
 
   if (dir_arg)
     {
-      for (; optind < argc; ++optind)
+      int i;
+      for (i = 0; i < n_files; i++)
 	{
 	  errors |=
-	    make_path (argv[optind], mode, mode, owner_id, group_id, 0, NULL);
+	    make_path (file[i], mode, mode, owner_id, group_id, 0,
+		       (x.verbose ? "creating directory `%s'" : NULL));
 	}
     }
   else
     {
-      if (optind == argc - 2)
-	{
-	  if (!isdir (argv[argc - 1]))
-	    errors = install_file_in_file (argv[argc - 2], argv[argc - 1]);
+      /* FIXME: it's a little gross that this initialization is
+	 required by copy.c::copy. */
+      hash_init (INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);
+
+      if (n_files == 2)
+        {
+          if (mkdir_and_install)
+	    errors = install_file_to_path (file[0], file[1], &x);
+	  else if (!isdir (file[1]))
+	    errors = install_file_in_file (file[0], file[1], &x);
 	  else
-	    errors = install_file_in_dir (argv[argc - 2], argv[argc - 1]);
+	    errors = install_file_in_dir (file[0], file[1], &x);
 	}
       else
 	{
-	  if (!isdir (argv[argc - 1]))
-	    usage (1);
-	  for (; optind < argc - 1; ++optind)
+	  int i;
+	  const char *dest = file[n_files - 1];
+	  if (!isdir (dest))
 	    {
-	      errors |= install_file_in_dir (argv[optind], argv[argc - 1]);
+	      error (0, 0,
+		     _("installing multiple files, but last argument (%s) \
+is not a directory"),
+		     dest);
+	      usage (1);
+	    }
+	  for (i = 0; i < n_files - 1; i++)
+	    {
+	      errors |= install_file_in_dir (file[i], dest, &x);
 	    }
 	}
     }
 
+  if (x.verbose)
+    close_stdout ();
   exit (errors);
 }
 
+/* Copy file FROM onto file TO, creating any missing parent directories of TO.
+   Return 0 if successful, 1 if an error occurs */
+
+static int
+install_file_to_path (const char *from, const char *to,
+		      const struct cp_options *x)
+{
+  char *dest_dir;
+  int fail;
+
+  dest_dir = dirname (to);
+
+  /* check to make sure this is a path (not install a b ) */
+  if (!STREQ (dest_dir, ".")
+      && !isdir (dest_dir))
+    {
+      /* Someone will probably ask for a new option or three to specify
+	 owner, group, and permissions for parent directories.  Remember
+	 that this option is intended mainly to help installers when the
+	 distribution doesn't provide proper install rules.  */
+#define DIR_MODE 0755
+      fail = make_path (dest_dir, DIR_MODE, DIR_MODE, owner_id, group_id, 0,
+			(x->verbose ? _("creating directory `%s'") : NULL));
+
+      if (fail == 0)
+	fail = install_file_in_dir (from, dest_dir, x);
+    }
+  else
+    {
+      fail = install_file_in_file (from, to, x);
+    }
+
+  free (dest_dir);
+
+  return fail;
+}
+
 /* Copy file FROM onto file TO and give TO the appropriate
    attributes.
    Return 0 if successful, 1 if an error occurs. */
 
 static int
-install_file_in_file (char *from, char *to)
+install_file_in_file (const char *from, const char *to,
+		      const struct cp_options *x)
 {
-  int to_created;
-  int no_need_to_chown;
-
-  if (copy_file (from, to, &to_created))
+  if (copy_file (from, to, x))
     return 1;
   if (strip_files)
     strip (to);
-  no_need_to_chown = (to_created
-		      && owner_name == NULL
-		      && group_name == NULL);
-  return change_attributes (to, no_need_to_chown);
+  if (change_attributes (to))
+    return 1;
+  if (x->preserve_timestamps)
+    return change_timestamps (from, to);
+  return 0;
 }
 
 /* Copy file FROM into directory TO_DIR, keeping its same name,
@@ -332,147 +462,49 @@ install_file_in_file (char *from, char *to)
    Return 0 if successful, 1 if not. */
 
 static int
-install_file_in_dir (char *from, char *to_dir)
+install_file_in_dir (const char *from, const char *to_dir,
+		     const struct cp_options *x)
 {
   char *from_base;
   char *to;
   int ret;
 
-  from_base = basename (from);
-  to = xmalloc ((unsigned) (strlen (to_dir) + strlen (from_base) + 2));
-  stpcpy (stpcpy (stpcpy (to, to_dir), "/"), from_base);
-  ret = install_file_in_file (from, to);
+  from_base = base_name (from);
+  to = path_concat (to_dir, from_base, NULL);
+  ret = install_file_in_file (from, to, x);
   free (to);
   return ret;
 }
 
-/* A chunk of a file being copied. */
-static char buffer[READ_SIZE];
-
 /* Copy file FROM onto file TO, creating TO if necessary.
-   Return 0 if the copy is successful, 1 if not.  If the copy is
-   successful, set *TO_CREATED to nonzero if TO was created (if it did
-   not exist or did, but was unlinked) and to zero otherwise.  If the
-   copy fails, don't modify *TO_CREATED.  */
+   Return 0 if the copy is successful, 1 if not.  */
 
 static int
-copy_file (char *from, char *to, int *to_created)
+copy_file (const char *from, const char *to, const struct cp_options *x)
 {
-  int fromfd, tofd;
-  int bytes;
-  int ret = 0;
-  struct stat from_stats, to_stats;
-  int target_created = 1;
-
-  if (stat (from, &from_stats))
-    {
-      error (0, errno, "%s", from);
-      return 1;
-    }
+  int fail;
+  int nonexistent_dst = 0;
+  int copy_into_self;
 
   /* Allow installing from non-regular files like /dev/null.
      Charles Karney reported that some Sun version of install allows that
      and that sendmail's installation process relies on the behavior.  */
-  if (S_ISDIR (from_stats.st_mode))
+  if (isdir (from))
     {
       error (0, 0, _("`%s' is a directory"), from);
       return 1;
     }
-  if (stat (to, &to_stats) == 0)
-    {
-      if (!S_ISREG (to_stats.st_mode))
-	{
-	  error (0, 0, _("`%s' is not a regular file"), to);
-	  return 1;
-	}
-      if (from_stats.st_dev == to_stats.st_dev
-	  && from_stats.st_ino == to_stats.st_ino)
-	{
-	  error (0, 0, _("`%s' and `%s' are the same file"), from, to);
-	  return 1;
-	}
 
-      /* The destination file exists.  Try to back it up if required.  */
-      if (backup_type != none)
-        {
-	  char *tmp_backup = find_backup_file_name (to);
-	  char *dst_backup;
+  fail = copy (from, to, nonexistent_dst, x, &copy_into_self, NULL);
 
-	  if (tmp_backup == NULL)
-	    error (1, 0, "virtual memory exhausted");
-	  dst_backup = (char *) alloca (strlen (tmp_backup) + 1);
-	  strcpy (dst_backup, tmp_backup);
-	  free (tmp_backup);
-	  if (rename (to, dst_backup))
-	    {
-	      if (errno != ENOENT)
-		{
-		  error (0, errno, "cannot backup `%s'", to);
-		  return 1;
-		}
-	    }
-	}
-
-      /* If unlink fails, try to proceed anyway.  */
-      if (unlink (to))
-	target_created = 0;
-    }
-
-  fromfd = open (from, O_RDONLY, 0);
-  if (fromfd == -1)
-    {
-      error (0, errno, "%s", from);
-      return 1;
-    }
-
-  /* Make sure to open the file in a mode that allows writing. */
-  tofd = open (to, O_WRONLY | O_CREAT | O_TRUNC, 0600);
-  if (tofd == -1)
-    {
-      error (0, errno, "%s", to);
-      close (fromfd);
-      return 1;
-    }
-
-  while ((bytes = safe_read (fromfd, buffer, READ_SIZE)) > 0)
-    if (full_write (tofd, buffer, bytes) < 0)
-      {
-	error (0, errno, "%s", to);
-	goto copy_error;
-      }
-
-  if (bytes == -1)
-    {
-      error (0, errno, "%s", from);
-      goto copy_error;
-    }
-
-  if (close (fromfd) < 0)
-    {
-      error (0, errno, "%s", from);
-      ret = 1;
-    }
-  if (close (tofd) < 0)
-    {
-      error (0, errno, "%s", to);
-      ret = 1;
-    }
-  if (ret == 0)
-    *to_created = target_created;
-  return ret;
-
- copy_error:
-  close (fromfd);
-  close (tofd);
-  return 1;
+  return fail;
 }
 
 /* Set the attributes of file or directory PATH.
-   If NO_NEED_TO_CHOWN is nonzero, don't call chown.
    Return 0 if successful, 1 if not. */
 
 static int
-change_attributes (char *path, int no_need_to_chown)
+change_attributes (const char *path)
 {
   int err = 0;
 
@@ -482,9 +514,6 @@ change_attributes (char *path, int no_need_to_chown)
      On System V, users can give away files with chown and then not
      be able to chmod them.  So don't give files away.
 
-     We don't pass -1 to chown to mean "don't change the value"
-     because SVR3 and earlier non-BSD systems don't support that.
-
      We don't normally ignore errors from chown because the idea of
      the install command is that the file is supposed to end up with
      precisely the attributes that the user specified (or defaulted).
@@ -492,7 +521,7 @@ change_attributes (char *path, int no_need_to_chown)
      want to know.  But AFS returns EPERM when you try to change a
      file's group; thus the kludge.  */
 
-  if (!no_need_to_chown && chown (path, owner_id, group_id)
+  if (chown (path, owner_id, group_id)
 #ifdef AFS
       && errno != EPERM
 #endif
@@ -508,6 +537,35 @@ change_attributes (char *path, int no_need_to_chown)
   return 0;
 }
 
+/* Set the timestamps of file TO to match those of file FROM.
+   Return 0 if successful, 1 if not. */
+
+static int
+change_timestamps (const char *from, const char *to)
+{
+  struct stat stb;
+  struct utimbuf utb;
+
+  if (stat (from, &stb))
+    {
+      error (0, errno, "%s", from);
+      return 1;
+    }
+
+  /* There's currently no interface to set file timestamps with
+     better than 1-second resolution, so discard any fractional
+     part of the source timestamp.  */
+
+  utb.actime = stb.st_atime;
+  utb.modtime = stb.st_mtime;
+  if (utime (to, &utb))
+    {
+      error (0, errno, "%s", to);
+      return 1;
+    }
+  return 0;
+}
+
 /* Strip the symbol table from the file PATH.
    We could dig the magic number out of the file first to
    determine whether to strip it, but the header files and
@@ -515,7 +573,7 @@ change_attributes (char *path, int no_need_to_chown)
    it portable would be very difficult.  Not worth the effort. */
 
 static void
-strip (char *path)
+strip (const char *path)
 {
   int pid, status;
 
@@ -526,7 +584,7 @@ strip (char *path)
       error (1, errno, _("fork system call failed"));
       break;
     case 0:			/* Child. */
-      execlp ("strip", "strip", path, (char *) NULL);
+      execlp ("strip", "strip", path, NULL);
       error (1, errno, _("cannot run strip"));
       break;
     default:			/* Parent. */
@@ -561,7 +619,7 @@ get_ids (void)
       endpwent ();
     }
   else
-    owner_id = getuid ();
+    owner_id = (uid_t) -1;
 
   if (group_name)
     {
@@ -570,7 +628,7 @@ get_ids (void)
 	{
 	  long int tmp_long;
 	  if (xstrtol (group_name, NULL, 0, &tmp_long, NULL) != LONGINT_OK
-	      || tmp_long < 0 || tmp_long > (long) GID_T_MAX)
+	      || tmp_long < 0 || tmp_long > GID_T_MAX)
 	    error (1, 0, _("invalid group `%s'"), group_name);
 	  group_id = (gid_t) tmp_long;
 	}
@@ -579,7 +637,7 @@ get_ids (void)
       endgrent ();
     }
   else
-    group_id = getgid ();
+    group_id = (gid_t) -1;
 }
 
 static void
@@ -597,18 +655,24 @@ Usage: %s [OPTION]... SOURCE DEST           (1st format)\n\
 "),
 	      program_name, program_name, program_name);
       printf (_("\
-In first two formats, copy SOURCE to DEST or multiple SOURCE(s) to\n\
-DIRECTORY, while setting permission modes and owner/group.  In third\n\
-format, make all components of the given DIRECTORY(ies).\n\
+In the first two formats, copy SOURCE to DEST or multiple SOURCE(s) to\n\
+the existing DIRECTORY, while setting permission modes and owner/group.\n\
+In the third format, create all components of the given DIRECTORY(ies).\n\
 \n\
   -b, --backup        make backup before removal\n\
   -c                  (ignored)\n\
-  -d, --directory     create [leading] directories, mandatory for 3rd format\n\
+  -d, --directory     treat all arguments as directory names; create all\n\
+                        components of the specified directories\n\
+  -D                  create all leading components of DEST except the last,\n\
+                        then copy SOURCE to DEST;  useful in the 1st format\n\
   -g, --group=GROUP   set group ownership, instead of process' current group\n\
-  -m, --mode=MODE     set permission mode (as in chmod), instead of rw-r--r--\n\
+  -m, --mode=MODE     set permission mode (as in chmod), instead of rwxr-xr-x\n\
   -o, --owner=OWNER   set ownership (super-user only)\n\
+  -p, --preserve-timestamps   apply access/modification times of SOURCE files\n\
+                        to corresponding destination files\n\
   -s, --strip         strip symbol tables, only for 1st and 2nd formats\n\
   -S, --suffix=SUFFIX override the usual backup suffix\n\
+      --verbose       print the name of each directory as it is created\n\
   -V, --version-control=WORD   override the usual version control\n\
       --help          display this help and exit\n\
       --version       output version information and exit\n\
@@ -622,7 +686,8 @@ version control may be set with VERSION_CONTROL, values are:\n\
   nil, existing   numbered if numbered backups exist, simple otherwise\n\
   never, simple   always make simple backups\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
