@@ -1,5 +1,5 @@
 /* `ln' program to create links between files.
-   Copyright (C) 86, 89, 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 86, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,6 +28,7 @@
 
 #include "system.h"
 #include "backupfile.h"
+#include "closeout.h"
 #include "error.h"
 
 int link ();			/* Some systems don't declare this anywhere. */
@@ -48,7 +49,7 @@ int symlink ();
 	tmp_source = (char *) alloca (strlen ((source)) + 1);		\
 	strcpy (tmp_source, (source));					\
 	strip_trailing_slashes (tmp_source);				\
-	source_base = basename (tmp_source);				\
+	source_base = base_name (tmp_source);				\
 									\
 	(new_dest) = (char *) alloca (strlen ((dest)) + 1		\
 				      + strlen (source_base) + 1);	\
@@ -56,17 +57,19 @@ int symlink ();
       }									\
     while (0)
 
-char *basename ();
+char *base_name ();
 char *dirname ();
 enum backup_type get_version ();
 int isdir ();
 int yesno ();
 void strip_trailing_slashes ();
-char *stpcpy ();
 
 /* The name by which the program was run, for error messages.  */
 char *program_name;
 
+/* FIXME: document */
+enum backup_type backup_type;
+
 /* A pointer to the function used to make links.  This will point to either
    `link' or `symlink'. */
 static int (*linkfunc) ();
@@ -105,13 +108,13 @@ static int show_version;
 static struct option const long_options[] =
 {
   {"backup", no_argument, NULL, 'b'},
-  {"directory", no_argument, &hard_dir_link, 1},
+  {"directory", no_argument, NULL, 'F'},
   {"no-dereference", no_argument, NULL, 'n'},
   {"force", no_argument, NULL, 'f'},
   {"interactive", no_argument, NULL, 'i'},
   {"suffix", required_argument, NULL, 'S'},
-  {"symbolic", no_argument, &symbolic_link, 1},
-  {"verbose", no_argument, &verbose, 1},
+  {"symbolic", no_argument, NULL, 's'},
+  {"verbose", no_argument, NULL, 'v'},
   {"version-control", required_argument, NULL, 'V'},
   {"help", no_argument, &show_help, 1},
   {"version", no_argument, &show_version, 1},
@@ -134,19 +137,23 @@ same_name (const char *source, const char *dest)
     error (1, 0, _("virtual memory exhausted"));
 
   if (stat (source_dirname, &source_dir_stats))
-    /* Shouldn't happen.  */
-    error (1, errno, "%s", source_dirname);
+    {
+      /* Shouldn't happen.  */
+      error (1, errno, "%s", source_dirname);
+    }
 
   if (stat (dest_dirname, &dest_dir_stats))
-    /* Shouldn't happen.  */
-    error (1, errno, "%s", dest_dirname);
+    {
+      /* Shouldn't happen.  */
+      error (1, errno, "%s", dest_dirname);
+    }
 
   free (source_dirname);
   free (dest_dirname);
 
   return (source_dir_stats.st_dev == dest_dir_stats.st_dev
 	  && source_dir_stats.st_ino == dest_dir_stats.st_ino
-	  && STREQ (basename (source), basename (dest)));
+	  && STREQ (base_name (source), base_name (dest)));
 }
 
 /* Make a link DEST to the (usually) existing file SOURCE.
@@ -181,13 +188,40 @@ do_link (const char *source, const char *dest)
     }
 
   lstat_status = lstat (dest, &dest_stats);
-
   if (lstat_status != 0 && errno != ENOENT)
     {
       error (0, errno, "%s", dest);
       return 1;
     }
 
+  /* If the destination is a directory or (it is a symlink to a directory
+     and the user has not specified --no-dereference), then form the
+     actual destination name by appending base_name (source) to the
+     specified destination directory.  */
+  if ((lstat_status == 0
+       && S_ISDIR (dest_stats.st_mode))
+#ifdef S_ISLNK
+      || (dereference_dest_dir_symlinks
+	  && (lstat_status == 0
+	      && S_ISLNK (dest_stats.st_mode)
+	      && isdir (dest)))
+#endif
+     )
+    {
+      /* Target is a directory; build the full filename. */
+      char *new_dest;
+      PATH_BASENAME_CONCAT (new_dest, dest, source);
+      dest = new_dest;
+
+      /* Get stats for new DEST.  */
+      lstat_status = lstat (dest, &dest_stats);
+      if (lstat_status != 0 && errno != ENOENT)
+	{
+	  error (0, errno, "%s", dest);
+	  return 1;
+	}
+    }
+
   /* If --force (-f) has been specified without --backup, then before
      making a link ln must remove the destination file if it exists.
      (with --backup, it just renames any existing destination file)
@@ -214,28 +248,6 @@ do_link (const char *source, const char *dest)
       return 1;
     }
 
-  /* If the destination is a directory or (it is a symlink to a directory
-     and the user has not specified --no-dereference), then form the
-     actual destination name by appending basename (source) to the
-     specified destination directory.  */
-  if ((lstat_status == 0
-       && S_ISDIR (dest_stats.st_mode))
-#ifdef S_ISLNK
-      || (dereference_dest_dir_symlinks
-	  && (S_ISLNK (dest_stats.st_mode)
-	  && isdir (dest)))
-#endif
-     )
-    {
-      /* Target is a directory; build the full filename. */
-      char *new_dest;
-      PATH_BASENAME_CONCAT (new_dest, dest, source);
-      dest = new_dest;
-      /* Set this to nonzero to force another call to lstat
-	 with the new destination.  */
-      lstat_status = 1;
-    }
-
   if (lstat_status == 0 || lstat (dest, &dest_stats) == 0)
     {
       if (S_ISDIR (dest_stats.st_mode))
@@ -249,7 +261,7 @@ do_link (const char *source, const char *dest)
 	  if (!yesno ())
 	    return 0;
 	}
-      else if (!remove_existing_files)
+      else if (!remove_existing_files && backup_type == none)
 	{
 	  error (0, 0, _("%s: File exists"), dest);
 	  return 1;
@@ -257,7 +269,7 @@ do_link (const char *source, const char *dest)
 
       if (backup_type != none)
 	{
-	  char *tmp_backup = find_backup_file_name (dest);
+	  char *tmp_backup = find_backup_file_name (dest, backup_type);
 	  if (tmp_backup == NULL)
 	    error (1, 0, _("virtual memory exhausted"));
 	  dest_backup = (char *) alloca (strlen (tmp_backup) + 1);
@@ -314,26 +326,28 @@ usage (int status)
   else
     {
       printf (_("\
-Usage: %s [OPTION]... SOURCE [DEST]\n\
-  or:  %s [OPTION]... SOURCE... DIRECTORY\n\
+Usage: %s [OPTION]... TARGET [LINK_NAME]\n\
+  or:  %s [OPTION]... TARGET... DIRECTORY\n\
 "),
 	      program_name, program_name);
       printf (_("\
-Link SOURCE to DEST (. by default), or multiple SOURCE(s) to DIRECTORY.\n\
-Makes hard links by default, symbolic links with -s.\n\
+Create a link to the specified TARGET with optional LINK_NAME.  If there is\n\
+more than one TARGET, the last argument must be a directory;  create links\n\
+in DIRECTORY to each TARGET.  Create hard links by default, symbolic links\n\
+with --symbolic.  When creating hard links, each TARGET must exist.\n\
 \n\
-  -b, --backup                 make backups for removed files\n\
-  -d, -F, --directory          hard link directories (super-user only)\n\
-  -f, --force                  remove existing destinations\n\
-  -n, --no-dereference         treat destination that is a symlink to a\n\
-                                 directory as if it were a normal file\n\
-  -i, --interactive            prompt whether to remove destinations\n\
-  -s, --symbolic               make symbolic links instead of hard links\n\
-  -S, --suffix=SUFFIX          override the usual backup suffix\n\
-  -v, --verbose                print name of each file before linking\n\
-  -V, --version-control=WORD   override the usual version control\n\
-      --help                   display this help and exit\n\
-      --version                output version information and exit\n\
+  -b, --backup                make a backup of each existing destination file\n\
+  -d, -F, --directory         hard link directories (super-user only)\n\
+  -f, --force                 remove existing destination files\n\
+  -n, --no-dereference        treat destination that is a symlink to a\n\
+                                directory as if it were a normal file\n\
+  -i, --interactive           prompt whether to remove destinations\n\
+  -s, --symbolic              make symbolic links instead of hard links\n\
+  -S, --suffix=SUFFIX         override the usual backup suffix\n\
+  -v, --verbose               print name of each file before linking\n\
+  -V, --version-control=WORD  override the usual version control\n\
+      --help                  display this help and exit\n\
+      --version               output version information and exit\n\
 \n\
 "));
       printf (_("\
@@ -344,7 +358,8 @@ version control may be set with VERSION_CONTROL, values are:\n\
   nil, existing   numbered if numbered backups exist, simple otherwise\n\
   never, simple   always make simple backups\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -371,9 +386,8 @@ main (int argc, char **argv)
     = hard_dir_link = 0;
   errors = 0;
 
-  while ((c = getopt_long (argc, argv,
-			   "bdfinsvFS:V:", long_options, (int *) 0))
-	 != EOF)
+  while ((c = getopt_long (argc, argv, "bdfinsvFS:V:", long_options, NULL))
+	 != -1)
     {
       switch (c)
 	{
@@ -422,6 +436,7 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("ln (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
@@ -434,8 +449,7 @@ main (int argc, char **argv)
       usage (1);
     }
 
-  if (make_backups)
-    backup_type = get_version (version);
+  backup_type = (make_backups ? get_version (version) : none);
 
 #ifdef S_ISLNK
   if (symbolic_link)
@@ -492,5 +506,7 @@ main (int argc, char **argv)
 	errors += do_link (argv[optind], to);
     }
 
+  if (verbose)
+    close_stdout ();
   exit (errors != 0);
 }
