@@ -1,5 +1,5 @@
 /* mkdir -- make directories
-   Copyright (C) 90, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 90, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 #include "system.h"
 #include "modechange.h"
 #include "makepath.h"
+#include "closeout.h"
 #include "error.h"
 
 /* The name this program was run with. */
@@ -42,7 +43,7 @@ static int show_version;
 static struct option const longopts[] =
 {
   {"mode", required_argument, NULL, 'm'},
-  {"parents", no_argument, &path_mode, 1},
+  {"parents", no_argument, NULL, 'p'},
   {"help", no_argument, &show_help, 1},
   {"verbose", no_argument, NULL, 2},
   {"version", no_argument, &show_version, 1},
@@ -67,7 +68,8 @@ Create the DIRECTORY(ies), if they do not already exist.\n\
       --help        display this help and exit\n\
       --version     output version information and exit\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -89,7 +91,7 @@ main (int argc, char **argv)
 
   path_mode = 0;
 
-  while ((optc = getopt_long (argc, argv, "pm:", longopts, (int *) 0)) != EOF)
+  while ((optc = getopt_long (argc, argv, "pm:", longopts, NULL)) != -1)
     {
       switch (optc)
 	{
@@ -112,6 +114,7 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("mkdir (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
