@@ -1,5 +1,5 @@
 /* mkfifo -- make fifo's (named pipes)
-   Copyright (C) 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@
 
 #include "system.h"
 #include "modechange.h"
+#include "closeout.h"
 #include "error.h"
 
 /* The name this program was run with. */
@@ -65,7 +66,8 @@ Create named pipes (FIFOs) with the given NAMEs.\n\
       --help        display this help and exit\n\
       --version     output version information and exit\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -90,7 +92,7 @@ main (int argc, char **argv)
 #ifndef S_ISFIFO
   error (4, 0, _("fifo files not supported"));
 #else
-  while ((optc = getopt_long (argc, argv, "m:", longopts, (int *) 0)) != EOF)
+  while ((optc = getopt_long (argc, argv, "m:", longopts, NULL)) != -1)
     {
       switch (optc)
 	{
@@ -107,6 +109,7 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("mkfifo (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
