@@ -1,5 +1,5 @@
 /* mvdir -- rename directory on System V r<4
-   Copyright (C) 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,15 +36,16 @@
 
 #include "system.h"
 #include "save-cwd.h"
+#include "closeout.h"
 #include "error.h"
 
 #ifndef HIPRI
-#define HIPRI -10
+# define HIPRI -10
 #endif
 
 #ifdef DEBUG
-#define link(FROM, TO) (printf("Linking %s to %s\n", FROM, TO), 0)
-#define unlink(FILE) (printf("Unlinking %s\n", FILE), 0)
+# define link(FROM, TO) (printf("Linking %s to %s\n", FROM, TO), 0)
+# define unlink(FILE) (printf("Unlinking %s\n", FILE), 0)
 #endif
 
 /* The name this program was run with. */
@@ -52,9 +53,8 @@ char *program_name;
 
 char *xgetcwd ();
 
-char *basename ();
+char *base_name ();
 char *dirname ();
-char *xmalloc ();
 void strip_trailing_slashes ();
 
 /* If nonzero, display usage information and exit.  */
@@ -84,6 +84,8 @@ Rename EXISTING_DIR to NEW_DIR.\n\
 \n\
    --help      display this help and exit\n\
    --version   output version information and exit\n"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -126,7 +128,7 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  while ((c = getopt_long (argc, argv, "", long_options, (int *) 0)) != EOF)
+  while ((c = getopt_long (argc, argv, "", long_options, NULL)) != -1)
     {
       switch (c)
 	{
@@ -140,7 +142,8 @@ main (int argc, char **argv)
 
   if (show_version)
     {
-      printf ("mvdir - %s\n", PACKAGE_VERSION);
+      printf ("mvdir (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
@@ -160,7 +163,7 @@ main (int argc, char **argv)
     error (2, 0, _("virtual memory exhausted"));
 
   /* Make sure `from' is not "." or "..". */
-  from_base = basename (from);
+  from_base = base_name (from);
   if (!strcmp (from_base, ".") || !strcmp (from_base, ".."))
     error (1, 0, _("cannot rename `.' or `..'"));
 
@@ -220,7 +223,7 @@ main (int argc, char **argv)
      and `to' starts with `../', we go to the full path of `to's parent,
      lest we try to reference the new directory's `..' while creating it.  */
 
-  to_base = basename (to);
+  to_base = base_name (to);
   i = strlen (to_base);
   to_dotdot = xmalloc (i + 4);
   strcpy (to_dotdot, to_base);
