@@ -1,5 +1,5 @@
 /* rmdir -- remove directories
-   Copyright (C) 90, 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,8 +28,17 @@
 #include <sys/types.h>
 
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
 
+#ifndef EEXIST
+# define EEXIST 0
+#endif
+
+#ifndef ENOTEMPTY
+# define ENOTEMPTY 0
+#endif
+
 void strip_trailing_slashes ();
 
 /* The name this program was run with. */
@@ -38,6 +47,13 @@ char *program_name;
 /* If nonzero, remove empty parent directories. */
 static int empty_paths;
 
+/* If nonzero, don't treat failure to remove a nonempty directory
+   as an error.  */
+static int ignore_fail_on_non_empty;
+
+/* If nonzero, output a diagnostic for every directory processed.  */
+static int verbose;
+
 /* If nonzero, display usage information and exit.  */
 static int show_help;
 
@@ -46,22 +62,29 @@ static int show_version;
 
 static struct option const longopts[] =
 {
-  {"path", no_argument, &empty_paths, 1},
-  {"parents", no_argument, &empty_paths, 1},
+  /* Don't name this `--force' because it's not close enough in meaning
+     to e.g. rm's -f option.  */
+  {"ignore-fail-on-non-empty", no_argument, NULL, CHAR_MAX + 1},
+
+  {"path", no_argument, NULL, 'p'},
+  {"parents", no_argument, NULL, 'p'},
+  {"verbose", no_argument, NULL, 14},
   {"help", no_argument, &show_help, 1},
   {"version", no_argument, &show_version, 1},
   {NULL, 0, NULL, 0}
 };
 
-/* Remove any empty parent directories of `path'.
-   Replaces '/' characters in `path' with NULs. */
+/* Remove any empty parent directories of PATH.
+   If PATH contains slash characters, at least one of them
+   (beginning with the rightmost) is replaced with a NUL byte.  */
 
-static void
+static int
 remove_parents (char *path)
 {
   char *slash;
+  int fail = 0;
 
-  do
+  while (1)
     {
       slash = strrchr (path, '/');
       if (slash == NULL)
@@ -71,8 +94,25 @@ remove_parents (char *path)
       while (slash > path && *slash == '/')
 	--slash;
       slash[1] = 0;
+
+      /* Give a diagnostic for each attempted removal if --verbose.  */
+      if (verbose)
+	error (0, errno, _("removing directory, %s"), path);
+
+      fail = rmdir (path);
+
+      if (fail)
+	{
+	  /* Give a diagnostic and set fail if not --ignore.  */
+	  if (!ignore_fail_on_non_empty || errno != ENOTEMPTY)
+	    {
+	      error (0, errno, "%s", path);
+	      fail = 1;
+	    }
+	  break;
+	}
     }
-  while (rmdir (path) == 0);
+  return fail;
 }
 
 static void
@@ -87,11 +127,16 @@ usage (int status)
       printf (_("\
 Remove the DIRECTORY(ies), if they are empty.\n\
 \n\
+      --ignore-fail-on-non-empty\n\
+                  ignore each failure that is solely because the\n\
+                  directory is non-empty\n\
   -p, --parents   remove explicit parent directories if being emptied\n\
+      --verbose   output a diagnostic for every directory processed\n\
       --help      display this help and exit\n\
       --version   output version information and exit\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -109,7 +154,7 @@ main (int argc, char **argv)
 
   empty_paths = 0;
 
-  while ((optc = getopt_long (argc, argv, "p", longopts, (int *) 0)) != EOF)
+  while ((optc = getopt_long (argc, argv, "p", longopts, NULL)) != -1)
     {
       switch (optc)
 	{
@@ -118,6 +163,12 @@ main (int argc, char **argv)
 	case 'p':
 	  empty_paths = 1;
 	  break;
+	case CHAR_MAX + 1:
+	  ignore_fail_on_non_empty = 1;
+	  break;
+	case 14:
+	  verbose = 1;
+	  break;
 	default:
 	  usage (1);
 	}
@@ -126,6 +177,7 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("rmdir (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
@@ -140,17 +192,33 @@ main (int argc, char **argv)
 
   for (; optind < argc; ++optind)
     {
+      int fail;
+      char *dir = argv[optind];
+
       /* Stripping slashes is harmless for rmdir;
 	 if the arg is not a directory, it will fail with ENOTDIR.  */
-      strip_trailing_slashes (argv[optind]);
-      if (rmdir (argv[optind]) != 0)
+      strip_trailing_slashes (dir);
+
+      /* Give a diagnostic for each attempted removal if --verbose.  */
+      if (verbose)
+	error (0, errno, _("removing directory, %s"), dir);
+
+      fail = rmdir (dir);
+
+      if (fail)
 	{
-	  error (0, errno, "%s", argv[optind]);
+	  if (ignore_fail_on_non_empty
+	      && (errno == ENOTEMPTY || errno == EEXIST))
+	    continue;
+
+	  error (0, errno, "%s", dir);
 	  errors = 1;
 	}
       else if (empty_paths)
-	remove_parents (argv[optind]);
+	{
+	  errors += remove_parents (dir);
+	}
     }
 
-  exit (errors);
+  exit (errors == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
