@@ -1,5 +1,5 @@
 /* touch -- change modification and access times of files
-   Copyright (C) 87, 89, 90, 91, 95, 1996 Free Software Foundation Inc.
+   Copyright (C) 87, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,21 +15,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-/* Options:
-   -a, --time={atime,access,use}	Change access time only.
-   -c, --no-create		Do not create files that do not exist.
-   -d, --date=TIME		Specify time and date in various formats.
-   -f				Ignored.
-   -m, --time={mtime,modify}	Change modification time only.
-   -r, --reference=FILE		Use the time and date of reference file FILE.
-   -t TIME			Specify time and date in the form
-				`MMDDhhmm[[CC]YY][.ss]'.
-
-   If no options are given, -am is the default, using the current time.
-   The -r, -t, and -d options are mutually exclusive.  If a file does not
-   exist, create it unless -c is given.
-
-   Written by Paul Rubin, Arnold Robbins, Jim Kingdon, David MacKenzie,
+/* Written by Paul Rubin, Arnold Robbins, Jim Kingdon, David MacKenzie,
    and Randy Smith. */
 
 #include <config.h>
@@ -38,23 +24,20 @@
 #include <sys/types.h>
 
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
 #include "argmatch.h"
+#include "getdate.h"
+#include "safe-read.h"
+#include "posixtm.h"
 
 #ifndef STDC_HEADERS
 time_t time ();
 #endif
 
-time_t get_date ();
-time_t posixtime ();
-int safe_read ();
 int full_write ();
 void invalid_arg ();
 
-#ifndef HAVE_UTIME_NULL
-static int utime_now ();
-#endif
-
 /* Bitmasks for `change_times'. */
 #define CH_ATIME 1
 #define CH_MTIME 2
@@ -99,7 +82,7 @@ static int show_version;
 
 static struct option const longopts[] =
 {
-  {"time", required_argument, 0, 130},
+  {"time", required_argument, 0, CHAR_MAX + 1},
   {"no-create", no_argument, 0, 'c'},
   {"date", required_argument, 0, 'd'},
   {"file", required_argument, 0, 'r'},
@@ -170,18 +153,18 @@ touch (char *file)
 
   if (amtime_now)
     {
-#ifndef HAVE_UTIME_NULL
-      status = utime_now (file, sbuf.st_size);
-#else
       /* Pass NULL to utime so it will not fail if we just have
 	 write access to the file, but don't own it.  */
       status = utime (file, NULL);
-#endif
     }
   else
     {
       struct utimbuf utb;
 
+      /* There's currently no interface to set file timestamps with
+	 better than 1-second resolution, so discard any fractional
+	 part of the source timestamp.  */
+
       if (use_ref)
 	{
 	  utb.actime = ref_stats.st_atime;
@@ -208,32 +191,6 @@ touch (char *file)
   return 0;
 }
 
-#ifndef HAVE_UTIME_NULL
-/* Emulate utime (file, NULL) for systems (like 4.3BSD) that do not
-   interpret it to set the access and modification times of FILE to
-   the current time.  FILESIZE is the correct size of FILE, used to
-   make sure empty files are not lengthened to 1 byte.
-   Return 0 if successful, -1 if not. */
-
-static int
-utime_now (const char *file, off_t filesize)
-{
-  int fd;
-  char c;
-  int status = 0;
-
-  fd = open (file, O_RDWR, 0666);
-  if (fd < 0
-      || safe_read (fd, &c, sizeof (char)) < 0
-      || lseek (fd, (off_t) 0, SEEK_SET) < 0
-      || full_write (fd, &c, sizeof (char)) < 0
-      || ftruncate (fd, filesize) < 0
-      || close (fd) < 0)
-    status = -1;
-  return status;
-}
-#endif
-
 static void
 usage (int status)
 {
@@ -243,6 +200,8 @@ usage (int status)
   else
     {
       printf (_("Usage: %s [OPTION]... FILE...\n"), program_name);
+      printf (_("  or : %s [-acm] MMDDhhmm[YY] FILE... (obsolescent)\n"),
+	      program_name);
       printf (_("\
 Update the access and modification times of each FILE to the current time.\n\
 \n\
@@ -252,14 +211,17 @@ Update the access and modification times of each FILE to the current time.\n\
   -f                     (ignored)\n\
   -m                     change only the modification time\n\
   -r, --reference=FILE   use this file's times instead of current time\n\
-  -t STAMP               use MMDDhhmm[[CC]YY][.ss] instead of current time\n\
+  -t STAMP               use [[CC]YY]MMDDhhmm[.ss] instead of current time\n\
       --time=WORD        access -a, atime -a, mtime -m, modify -m, use -a\n\
       --help             display this help and exit\n\
       --version          output version information and exit\n\
 \n\
 STAMP may be used without -t if none of -drt, nor --, are used.\n\
+Note that the three time-date formats recognized for the -d and -t options\n\
+and for the obsolescent argument are all different.\n\
 "));
-      puts (_("\nReport bugs to fileutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
+      close_stdout ();
     }
   exit (status);
 }
@@ -279,8 +241,7 @@ main (int argc, char **argv)
   change_times = no_create = use_ref = posix_date = flexible_date = 0;
   newtime = (time_t) -1;
 
-  while ((c = getopt_long (argc, argv, "acd:fmr:t:", longopts, (int *) 0))
-	 != EOF)
+  while ((c = getopt_long (argc, argv, "acd:fmr:t:", longopts, NULL)) != -1)
     {
       switch (c)
 	{
@@ -317,13 +278,14 @@ main (int argc, char **argv)
 
 	case 't':
 	  posix_date++;
-	  newtime = posixtime (optarg);
+	  newtime = posixtime (optarg,
+			       PDS_LEADING_YEAR | PDS_CENTURY | PDS_SECONDS);
 	  if (newtime == (time_t) -1)
 	    error (1, 0, _("invalid date format `%s'"), optarg);
 	  date_set++;
 	  break;
 
-	case 130:
+	case CHAR_MAX + 1:
 	  i = argmatch (optarg, time_args);
 	  if (i < 0)
 	    {
@@ -341,6 +303,7 @@ main (int argc, char **argv)
   if (show_version)
     {
       printf ("touch (%s) %s\n", GNU_PACKAGE, VERSION);
+      close_stdout ();
       exit (0);
     }
 
@@ -364,9 +327,9 @@ main (int argc, char **argv)
       date_set++;
     }
 
-  if (!date_set && optind < argc && strcmp (argv[optind - 1], "--"))
+  if (!date_set && optind < argc && !STREQ (argv[optind - 1], "--"))
     {
-      newtime = posixtime (argv[optind]);
+      newtime = posixtime (argv[optind], PDS_TRAILING_YEAR);
       if (newtime != (time_t) -1)
 	{
 	  optind++;
